import hcat.lib.functional
from hcat.lib.utils import calculate_indexes, load, cochlea_to_xml, correct_pixel_size_image, \
    rescale_box_sizes, normalize_image, make_rgb, image_to_float, save_image_as_png
from hcat.lib.cell import Cell
from hcat.lib.cochlea import Cochlea
from hcat.backends.detection import FasterRCNN_from_url
from hcat.lib.utils import warn
from torchvision.transforms.functional import gaussian_blur

import torch
from torch import Tensor
from torchvision.models.detection import FasterRCNN
from tqdm import tqdm
from itertools import product
import numpy as np
from hcat.lib.explore_lif import get_xml
import torchvision.ops
import skimage.io as io
from numpy.typing import ArrayLike

import os.path
from typing import Optional, List, Dict, Callable, Tuple, Union

__model_url__ = 'https://www.dropbox.com/s/opf43jwcbgz02vm/detection_trained_model.trch?dl=1'

@torch.no_grad()
def _detect(f: Optional[str] = None,
            image_base: Optional[Tensor] = None,
            curve_path: str = None,
            cell_detection_threshold: float = 0.57,
            nms_threshold: float = 0.1,
            dtype: Optional[bool] = None,
            model: Optional[FasterRCNN] = None,
            scale: Optional[int] = None,
            save_xml: Optional[bool] = False,
            save_fig: Optional[bool] = False,
            save_png: Optional[bool] = False,
            save_csv: Optional[bool] = False,
            normalize: Optional[bool] = False,
            pixel_size: Optional[Union[bool, int, float]] = None,
            cell_diameter: Optional[Union[bool, int, float]] = None,
            return_model: Optional[bool] = False,
            no_curve: Optional[bool] = False,
            verbose: Optional[bool] = True):
    """
    2D hair cell detection algorithm.
    Loads arbitrarily large 2d image and performs iterative faster rcnn detection on the entire image.

    :param *str* f: path to image by which to analyze
    :param *float* cell_detection_threshold: cells below threshold are rejected
    :param *float* nms_threshold: iou rejection threshold for nms.
    :return: *Cochlea* object containing data of analysis.
    """

    if verbose: print(f'Performing Hair Cell Detection Task')
    if f is None and image_base is None:
        warn('ERROR: No File to Analyze... \nAborting.', color='red')
        return None
    if not (pixel_size or cell_diameter) and verbose:
        warn('WARNING: Pixel Size is not set. Defaults to 288.88 nm x/y. '
             'Consider supplying value for optimal performance.', color='yellow')

    device = hcat.lib.utils.get_device(verbose=verbose)

    if image_base is None and f:
        # Load and preprocess Image
        image_base: np.ndarray = load(f, 'reMerged', verbose=verbose)  # from hcat.lib.utils
        image_base: np.ndarray = image_base[:, ...].max(-1) if image_base.ndim == 4 else image_base

        dtype = image_base.dtype if dtype is None else dtype
        scale: int = scale if scale else hcat.lib.utils.get_dtype_offset(dtype, image_base.max())

        image_base: Tensor = image_to_float(image_base, scale, verbose=verbose)
        image_base: Tensor = image_base.to(device)
        image_base: Tensor = correct_pixel_size_image(image_base, pixel_size, cell_diameter, verbose=verbose)
        image_base: Tensor = make_rgb(image_base)
        image_base: Tensor = normalize_image(image_base, normalize, verbose=True)

    elif image_base is not None:
        image_base = image_base.to(device)

    if save_png:
        save_image_as_png(image_base, filename=f, verbose=verbose)


    # Initalize the model...
    if model is None:
        print(f'[      ] Initializing detection model...', end='')
        model: FasterRCNN = FasterRCNN_from_url(url=__model_url__, device=device)
        print("\r[\x1b[1;32;40m DONE \x1b[0m]")
    else:
        model = model.to(device)

    if device == 'mps':
        model = model.to(torch.float32)

    prepare_image: Callable = lambda x: x.to(torch.float32)  # x.sub(0.5).div(0.5).float()

    # Initalize curvature detection
    predict_curvature = hcat.lib.functional.PredictCurvature(erode=3)  # Callable[[Tensor], Tuple[Tensor, Tensor]]

    # Get the indicies for evaluating cropped regions
    _, x, y = image_base.shape
    if verbose: print(image_base.shape)
    x_ind: List[List[int]] = calculate_indexes(10, 235, x, x)  # [[0, 255], [30, 285], ...]
    y_ind: List[List[int]] = calculate_indexes(10, 235, y, y)  # [[0, 255], [30, 285], ...]
    total: int = len(x_ind) * len(y_ind)

    # Initalize other small things
    cell_id = 1
    cells = []
    add_cell: Callable = cells.append  # stupid but done for speed

    generator = tqdm(product(x_ind, y_ind), total=total, desc='Detecting: ') if verbose else product(x_ind, y_ind)

    for x, y in generator:

        # Load and prepare image crop for ML model evaluation
        image: Tensor = image_base[:, x[0]:x[1], y[0]:y[1]]

        if image.max() == 0 or image.min() == image.max():
            continue

        image: List[Tensor] = [prepare_image(image).to(torch.float32)]

        # Evaluate Deep Learning Model
        out = model(image)

        out = out[1][0] if isinstance(out, tuple) else out[0]

        ind: Tensor = out['scores'] > cell_detection_threshold

        scores: Tensor = out['scores'][ind].cpu()
        boxes: Tensor = out['boxes'][ind, :]
        labels: Tensor = out['labels'][ind]

        # The model output coords with respect to the crop of image_base. We have to adjust
        # Format of torch images expect to be [B, C, Height, Width]
        boxes[:, [0, 2]] += y[0]
        boxes[:, [1, 3]] += x[0]
        # boxes = rescale_box_sizes(boxes, current_pixel_size=pixel_size, cell_diameter=cell_diameter)

        # center x, center y, width, height
        centers: Tensor = torchvision.ops.box_convert(boxes, 'xyxy', 'cxcywh').cpu()
        cx = centers[:, 0]
        cy = centers[:, 1]

        for i, score in enumerate(scores):
            add_cell(Cell(id=cell_id,
                          loc=torch.tensor([0, cx[i], cy[i], 0]),
                          image=None,
                          mask=None,
                          cell_type='OHC' if labels[i] == 1 else 'IHC',
                          boxes=boxes[i, :],
                          scores=scores[i]))
            cell_id += 1

    # some cells may overlap. We remove cells after analysis is complete.
    if nms_threshold < 1:
        cells: List[Cell] = _cell_nms(cells, nms_threshold)

    ohc = sum([int(c.type == 'OHC') for c in cells])  # number of ohc
    ihc = sum([int(c.type == 'IHC') for c in cells])  # number of ihc
    if verbose:
        print(f'Total Cells: {len(cells)}\n   OHC: {ohc}\n   IHC: {ihc}')

    if not no_curve:
        max_projection: Tensor = image_base[[1], ...].mul(0.5).add(0.5).unsqueeze(-1).cpu()
        curvature, distance, apex = predict_curvature(max_projection, cells, curve_path)

        if curvature is None and verbose:
            warn('WARNING: All three methods to predict hair cell path have failed. Frequency Mapping functionality is '
                 'limited. Consider Manual Calculation.', color='yellow')

        # curvature estimation really only works if there is a lot of tissue...
        if distance is not None and distance.max() > 4000:
            for c in cells: c.calculate_frequency(curvature[[0, 1], :], distance)  # calculate cell's best frequency
            cells = [c for c in cells if not c._distance_is_far_away]  # remove a cell if its far away from curve

        else:
            curvature, distance, apex = None, None, None
            if verbose:
                warn('WARNING: Predicted Cochlear Distance is below 4000um. Not sufficient '
                     'information to determine cell frequency.', color='yellow')
    else:
        curvature, distance, apex = None, None, None

    xml = get_xml(f) if f and f.endswith('.lif') else None
    filename = os.path.split(f)[-1] if f else None

    # remove weird cell ID's
    for i, c in enumerate(cells): c.id = i + 1

    # Store in compressible object for further use
    c = Cochlea(mask=None,
                filename=filename,
                path=f,
                analysis_type='detect',
                leica_metadata=xml,
                im_shape=image_base.shape,
                cochlear_distance=distance,
                curvature=curvature,
                cells=cells,
                apex=apex)

    if save_csv: c.write_csv()
    if save_xml: cochlea_to_xml(c)
    if save_fig: c.make_detect_fig(image_base)

    if return_model:
        return c, model
    else:
        return c


def _cell_nms(cells: List[Cell], nms_threshold: float) -> List[Cell]:
    """
    Perforns non maximum supression on the resulting cell predictions

    :param cells: Iterable of cells
    :param nms_threshold: cell iou threshold
    :return: Iterable of cells
    """
    # nms to get rid of cells
    boxes = torch.zeros((len(cells), 4))
    scores = torch.zeros(len(cells))
    for i, c in enumerate(cells):
        boxes[i, :] = c.boxes
        scores[i] = c.scores

    ind = torchvision.ops.nms(boxes, scores, nms_threshold)

    # need to pop off list elements from an int64 tensor
    ind_bool = torch.zeros(len(cells))
    ind_bool[ind] = 1
    for i, val in enumerate(ind_bool):
        if val == 0:
            cells[i] = None

    return [c for c in cells if c]



import torch
import torch.nn as nn
from src.models.depreciated.HCBlock import HCBlock
from warnings import filterwarnings
from typing import List
from src.train.transforms import _crop


filterwarnings("ignore", category=UserWarning)


class _HCNet(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, complexity: int = 30):
        super(HCNet, self).__init__()

        self.conv3x3_1 = nn.Conv3d(in_channels=in_channels, out_channels=10, padding=1, kernel_size=3)
        self.bn_1 = nn.BatchNorm3d(10)

        self.strided_conv = nn.Conv3d(10, complexity, kernel_size=3, stride=2, padding=1)
        self.bn_strided = nn.BatchNorm3d(complexity)

        self.hcblock = torch.jit.script(HCBlock(in_channels=complexity))

        self.transposed_conv = nn.ConvTranspose3d(in_channels=complexity, out_channels=complexity,
                                                  stride=(2, 2, 2), kernel_size=(3, 3, 3), padding=(1, 1, 1))
        self.batch_norm_transpose = nn.BatchNorm3d(complexity)

        self.conv5x5_1 = nn.Conv3d(in_channels=complexity, out_channels=complexity*2, kernel_size=5, padding=2)
        self.bn_5x5_1 = nn.BatchNorm3d(complexity*2)

        self.conv3x3_2 = nn.Conv3d(complexity*2, complexity*3, kernel_size=3, padding=1)
        self.bn_2 = nn.BatchNorm3d(complexity*3)

        self.out_conv = nn.Conv3d(complexity*3, out_channels=out_channels, kernel_size=1, padding=0)

        self.activation = nn.LeakyReLU()
        self.tanh = nn.Tanh()

    def forward(self, x: torch.Tensor, i: int = 5) -> torch.Tensor:
        x = self.activation(self.bn_1(self.conv3x3_1(x)))
        x = self.activation(self.bn_strided(self.strided_conv(x)))
        y = torch.zeros(x.shape).to(x.device)

        for t in range(i):
            in_ = torch.cat((x, y), dim=1)
            y = self.hcblock(in_) + y

        y = self.activation(self.batch_norm_transpose(self.transposed_conv(y)))
        y = self.activation(self.bn_5x5_1(self.conv5x5_1(y)))
        y = self.activation(self.bn_2(self.conv3x3_2(y)))
        y = self.out_conv(y)

        y = self.tanh(y) # EXPERIMENTAL

        return y

class HCNet(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, complexity: int = 30):
        super(HCNet, self).__init__()

        self.conv3x3_1 = nn.Conv3d(in_channels=in_channels, out_channels=10, padding=3, kernel_size=7)
        self.bn_1 = nn.BatchNorm3d(10)

        self.strided_conv = nn.Conv3d(10, complexity, kernel_size=5, stride=(5, 5, 3), padding=1)
        self.bn_strided = nn.BatchNorm3d(complexity)

        self.hcblock = torch.jit.script(HCBlock(in_channels=complexity))

        self.transposed_conv = nn.ConvTranspose3d(in_channels=complexity, out_channels=complexity,
                                                  stride=(5, 5, 3), kernel_size=5, padding=(1, 1, 1))

        self.batch_norm_transpose = nn.BatchNorm3d(complexity)

        # Embedd!!
        self.conv5x5_e = nn.Conv3d(in_channels=10 + complexity, out_channels=complexity * 2, kernel_size=5, padding=2)
        self.bn_5x5_e = nn.BatchNorm3d(complexity * 2)

        self.conv3x3_e = nn.Conv3d(complexity * 2, complexity * 3, kernel_size=3, padding=1)
        self.bn_e = nn.BatchNorm3d(complexity * 3)

        self.out_conv_e = nn.Conv3d(complexity * 3, out_channels=3, kernel_size=3, padding=1)

        # Prob!!!
        self.conv5x5_p = nn.Conv3d(in_channels=10 + complexity, out_channels=complexity * 2, kernel_size=5, padding=2)
        self.bn_5x5_p = nn.BatchNorm3d(complexity * 2)

        self.conv3x3_p = nn.Conv3d(complexity * 2, complexity * 3, kernel_size=3, padding=1)
        self.bn_p = nn.BatchNorm3d(complexity * 3)

        self.out_conv_p = nn.Conv3d(complexity * 3, out_channels=1, kernel_size=3, padding=1)

        self.activation = nn.LeakyReLU()
        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()

    def forward(self, x: torch.Tensor, i: int = 5) -> torch.Tensor:
        x = self.activation(self.bn_1(self.conv3x3_1(x)))
        y = self.activation(self.bn_strided(self.strided_conv(x)))
        z = torch.zeros(y.shape, device=y.device)

        for _ in range(i):
            z = self.activation(self.hcblock(torch.cat((y, z), dim=1)) + y)

        z = self.activation(self.batch_norm_transpose(self.transposed_conv(z)))

        z = torch.cat((crop(x, z.shape), z), dim=1)  # Skip Connection to add fine detail

        # embed
        e = self.activation(self.bn_5x5_e(self.conv5x5_e(z)))
        e = self.activation(self.bn_e(self.conv3x3_e(e)))
        e = self.out_conv_e(e)

        # prob
        p = self.activation(self.bn_5x5_p(self.conv5x5_p(z)))
        p = self.activation(self.bn_p(self.conv3x3_p(p)))
        p = self.out_conv_p(p)

        return torch.cat((self.tanh(e), self.sigmoid(p)), dim=1) # embed and prob

@torch.jit.script
def crop(x: torch.Tensor, shape: List[int]) -> torch.Tensor:
    return _crop(x, 0, 0, 0, int(shape[2]), int(shape[3]), int(shape[4]))



import pandas
from sklearn.cluster import DBSCAN
import torch
from torch import Tensor
from hcat.train.transforms import erosion
import hcat.lib.utils
from hcat import ShapeError

from typing import Tuple, Dict, Optional, List

import numpy as np
import skimage
import skimage.exposure
import skimage.filters
import skimage.morphology
import skimage.feature
import skimage.segmentation
import skimage.transform
import skimage.feature

from hcat.train.transforms import _crop
from hcat.lib.cell import Cell
from hcat.lib.utils import graceful_exit

import scipy.ndimage
import scipy.ndimage.morphology
from scipy.interpolate import splprep, splev
import GPy

import torchvision.ops
import torch.nn as nn

import matplotlib.pyplot as plt


class VectorToEmbedding(nn.Module):
    def __init__(self, scale: Tensor = torch.tensor([50]), device='cuda'):
        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor([scale])
        self.scale = scale.to(device) if scale.numel() != 1 else torch.tensor([scale, scale, scale],
                                                                              device=device).float()
        super(VectorToEmbedding, self).__init__()

    def forward(self, vector: Tensor, n: int = 1) -> Tensor:
        """
        :param vector: [B, 3=[x,y,z], X, Y, Z] prediction vector from spatial embedding model
        :return: [B, 3=[x,y,z], X, Y, Z] projected spatial embedding vector
        """
        if vector.ndim != 5: raise RuntimeError('Expected input tensor ndim == 5')

        num: Tensor = self.scale.float()

        with torch.no_grad():
            _x = torch.linspace(0, vector.shape[2] - 1, vector.shape[2], device=vector.device)
            _y = torch.linspace(0, vector.shape[3] - 1, vector.shape[3], device=vector.device)
            _z = torch.linspace(0, vector.shape[4] - 1, vector.shape[4], device=vector.device)

            xv, yv, zv = torch.meshgrid([_x, _y, _z])

            mesh = torch.cat((xv.unsqueeze(0).unsqueeze(0),
                              yv.unsqueeze(0).unsqueeze(0),
                              zv.unsqueeze(0).unsqueeze(0)), dim=1)

        # Apply the mesh grid to the vector! We always do this at least once!
        vector = vector.mul(num.view(1, 3, 1, 1, 1))
        mesh = mesh + vector

        for _ in range(n - 1):  # Only executes if n > 1
            index = mesh.round()
            b, c, x, y, z = index.shape
            for i, k in enumerate([x, y, z]):
                index[:, i, ...] = torch.clamp(index[:, i, ...], 0, k)

            index = (index[:, [0], ...] * y * z) + (index[:, [1], ...] * z) + (index[:, [2], ...])
            out_of_bounds = torch.logical_and(index > x * y * z - 1, index < 0)
            index[out_of_bounds] = 0

            # x_ind  = index[:,0,...].flatten()
            # y_ind  = index[:,1,...].flatten()
            # z_ind  = index[:,2,...].flatten()

            index = index.clamp(0, x * y * z - 1).long()
            # mesh = mesh + vector.flatten(start_dim=2)[]
            for i, shape in enumerate((x,y,z)):
                mesh[:, [i], ...] = mesh[:, [i], ...] + vector[:, [i], ...].take(index)
                out_of_bounds = torch.logical_or(mesh[:, [i], ...] > shape, mesh < 0)
                mesh[out_of_bounds] = -1

        return mesh


class EmbeddingToProbability(nn.Module):
    def __init__(self, scale: Tensor = torch.tensor([25]), device='cuda'):
        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor([scale])
        self.scale = scale.to(device) if scale.numel() != 1 else torch.tensor([scale, scale, scale], device=device)
        super(EmbeddingToProbability, self).__init__()

    def forward(self, embedding: Tensor, centroids: Tensor, sigma: Tensor) -> Tensor:
        """
        Calculates the euclidean distance between the centroid and the embedding
        embedding [B, 3, X, Y, Z] -> euclidean_norm[B, 1, X, Y, Z]
        euclidean_norm = sqrt(Δx^2 + Δy^2 + Δz^2) where Δx = (x_embed - x_centroid_i)

                             /    (e_ix - C_kx)^2       (e_iy - C_ky)^2        (e_iz - C_kz)^2   \
          prob_k(e_i) = exp |-1 * ----------------  -  -----------------   -  ------------------  |
                            \     2*sigma_kx ^2         2*sigma_ky ^2          2 * sigma_kz ^2  /


        Example:

        >>> from hcat.lib.functional import EmbeddingToProbability
        >>> import torch
        >>> embed = torch.load('embed.trch') # [B, C, X, Y, Z]
        >>> emb2prob = torch.jit.script(EmbeddingToProbability)
        >>> probability = emb2prob(embed)


        :param embedding: [B, K=3, X, Y, Z] embedding tensor where K is the likely centroid component: {X, Y, Z}
        :param centroids: [B, N, K_true=3] object centroids where N is the number of instances in the image and K_true is centroid {x, y, z}
        :param sigma: Tensor of shape = (1) or (embedding.shape)
        :return: [B, N, X, Y, Z] of probabilities for instance N
        """
        num = self.scale

        b, _, x, y, z = embedding.shape
        _, n, _ = centroids.shape

        # Centroids might be empty. In this case, return empty array!
        if n == 0:
            return torch.zeros((b, n, x, y, z), device=embedding.device)

        sigma = sigma + torch.tensor([1e-10], device=centroids.device)  # when sigma goes to zero, things tend to break

        if sigma.numel() == 1:
            sigma = torch.cat((sigma, sigma, sigma), dim=0)

        # Sometimes centroids might be in pixel coords instead of scaled.
        # If so, scale by num (usually 512)
        # centroids = centroids / num if centroids.max().gt(5) else centroids

        b, _, x, y, z = embedding.shape
        _, n, _ = centroids.shape
        # prob = torch.zeros((b, n, x, y, z), device=embedding.device)
        prob_list = torch.jit.annotate(List[Tensor], [])

        # Common operation. Done outside of loop for speed.
        newshape = (centroids.shape[0], 3, 1, 1, 1)
        sigma = sigma.pow(2).mul(2).view(newshape).mul(-1)

        # Calculate euclidean distance between centroid and embedding for each pixel and
        # turn that distance to probability and put it in preallocated matrix for each n
        # In eval mode uses in place operations to save memory!

        # A bit scuffed but should optimize well in torchscript
        for i in range(n):
            prob_list += [torch.exp(
                (embedding - centroids[:, i, :].view(newshape)).pow(2)
                                                               .div(sigma)
                                                               .sum(dim=1)
            ).squeeze(1)]
        prob = torch.stack(prob_list, dim=1)

        # else:  # If in Eval Mode - do a bunch of inplace operations for speed and memory efficiency
        #     for i in range(centroids.shape[1]):
        #         euclidean_norm = (embedding - centroids[:, i, :].view(centroids.shape[0], 3, 1, 1, 1)).pow(2)
        #         prob[:, i, :, :, :] = torch.exp(euclidean_norm.div_(sigma).sum(dim=1)).squeeze(1)
        return prob


@torch.jit.script
def learnable_centroid(embedding: Tensor, mask: Tensor, method: str = 'mean',
                       scale: Tensor = torch.tensor([25])) -> Tensor:
    """
    Estimates a center of attraction best predicted by the network.

    Take mean direction of vectors predicted for a single object and takes that as the centroid of the object.
    In this way the model can learn where best to predict a center.

    Example:

    >>> from hcat.lib.functional import learnable_centroid
    >>> import torch
    >>> embed = torch.load('embed.trch') # [B, C, X, Y, Z] predicted embedding
    >>> mask = torch.load('mask.trch') # [B, C, X, Y, Z] ground truth mask
    >>> centroids = learnable_centroid(embed)

    :param embedding: [B, C, X, Y, Z] predicted embedding tensor from deep learning model
    :param mask: [B, C, X, Y, Z] ground truth mask
    :param method: ['mode' or 'mean'] method by which to calculate centroids. Either can calculate the centroid of via "median" or "mean" of predicted embeddings
    :return: [C, N, k=3] centroids for N objects
    """
    with torch.no_grad():
        if method != 'mean' and method != 'mode':
            raise NotImplementedError(f'Cannot estimate centroids with method {method}')

        shape = embedding.shape
        mask = _crop(mask, 0, 0, 0, shape[2], shape[3], shape[4])  # from hcat.transforms._crop
        centroid = torch.zeros((mask.shape[0], mask.shape[1], 3), device=mask.device)

        # Loop over each instance and take the mean of the vectors multiplied by the mask (0 or 1)
        for i in range(mask.shape[1]):
            ind = torch.nonzero(embedding * mask[:, i, ...].unsqueeze(1))  # [:, [B, N, X, Y, Z]]
            center = torch.tensor([-10, -10, -10], device=mask.device)

            # Guess Centroids by average location of all vectors or just by the most common value (mean vs mode).
            if method == 'mean' and ind.shape[0] > 1:
                center = embedding[ind[:, 0], :, ind[:, 2], ind[:, 3], ind[:, 4]].mean(0)
            elif method == 'mode' and ind.shape[0] > 1:
                center = embedding[ind[:, 0], :, ind[:, 2], ind[:, 3], ind[:, 4]].mul(scale).round().mode(0)[0].div(scale)  # mode gives vals and indicies

            centroid[:, i, :] = center

        return centroid * scale.to(centroid.device)


class EstimateCentroids(nn.Module):
    def __init__(self,
                 scale: Tensor,
                 downsample: int = 2,
                 n_erode: int = 1,
                 eps: float = 0.5,
                 min_samples: int = 10,
                 device='cpu'):
        """
        Estimates centroids of objects in a spatial embedding matrix.

        :param downsample: factor by which to downsample embedding marix, higher numbers increase computational speed.
        :param n_erode: Erodes probability map, higher numbers prefentiall select voxels near object centers.
        :param eps: tuning parameter for DBSCAN
        :param min_samples: minimum number of voxels for an image to be detected
        :param scale: scaling factor of vector embedding. Must be identical to VectorToEmbedding
        :param device: physical device by which to perform computation on.
        """

        super(EstimateCentroids, self).__init__()
        self.downsample = downsample
        self.n_erode = n_erode
        self.eps = eps
        self.min_samples = min_samples
        if not isinstance(scale, torch.Tensor):
            scale = torch.tensor([scale])
        self.scale = scale.to(device) if scale.numel != 1 else torch.tensor([scale, scale, scale], device=device)

    def forward(self, embedding: Tensor, probability_map: Tensor = None,
                ) -> Tensor:
        """
        Estimates centroids of objects in a spatial embedding matrix.

        Employs DBSCAN clustering algorithm to predict embedding clusters in a predicted embedding matrix

        Example:

        >>> from hcat.lib.functional import EstimateCentroids
        >>> import torch
        >>> embed = torch.load('embed.trch') # [B, C, X, Y, Z] predicted embedding
        >>> mask = torch.load('mask.trch') # [B, C, X, Y, Z] ground truth mask
        >>> estimate = EstimateCentroids()
        >>> centroids = estimate(embed, mask)

        :param embedding: [B, C=3, X, Y, Z] predicted spatial embedding matrix
        :param probability_map: [B, C=1, X, Y, Z] predicted object probability mask
        :return: [B, N, K=3] Predicted centroids
        """


        # num = self.scale

        device = embedding.device
        embedding = embedding.squeeze(0).reshape((3, -1)).clone()

        binary_erosion = erosion(device=device)

        # First Step is to prune number of viable points
        # this is done optionally by passing the probability map of the image
        # Runs a binary erosion to only consider pixels in the middle of cells
        if probability_map is not None:
            if probability_map.ndim != 5: raise ShapeError(f'probability_map.ndim != 5, {probability_map.ndim}')
            if probability_map.shape[0] != 1: raise ShapeError(f'Batch size should be 1, not {probability_map.shape}')
            if probability_map.shape[1] != 1: raise ShapeError(f'Color size should be 1, not {probability_map.shape}')

            pm = probability_map.squeeze(0).gt(0.8).float()  # Needs 4 channel input. Remove Batch.
            for i in range(self.n_erode):
                pm = binary_erosion(pm)

            embedding = embedding[:, pm.squeeze(0).gt(0.5).view(-1)]

        # x = embedding[0, :]
        # y = embedding[1, :]
        # z = embedding[2, :]
        #
        # ind_x = x >= 0
        # ind_y = y >= 0
        # ind_z = z >= 0
        # ind = torch.logical_and(ind_x, ind_y)
        # ind = torch.logical_and(ind, ind_z)
        # embedding = embedding[:, ind]

        # Downsample the embeddings if there are too many points...
        embedding = embedding[:, 0:-1:self.downsample]

        # Remove obviously wrong embeddings...
        ind = torch.logical_not(embedding.le(1).sum(0).gt(1))  # any x,y,z pair with a negativ enumber is bad
        embedding = embedding[:, ind].round().cpu().detach()


        if embedding.numel() == 0:  # nothing detected
            return torch.empty((1, 0, 3), device=device)

        if embedding.shape[-1] > 3_000_000:
            raise RuntimeError(f'Unsuported number of samples. {embedding.shape[-1]} after pruning, 3,000,000 maximum.')

        if embedding.shape[1] == 0:
            return torch.empty((1, 0, 3), device=device)

        db = DBSCAN(eps=self.eps, min_samples=self.min_samples, n_jobs=-1).fit(embedding.numpy().T)
        labels = torch.from_numpy(db.labels_)

        unique, counts = torch.unique(labels, return_counts=True)

        centroids = torch.zeros((1, 3, unique.shape[0]), device=device)  # [B, C=3, N]
        scores = torch.zeros((1, 1, unique.shape[0]), device=device)  # [B, C=1, N]

        # Assign mean value of labels to the centroid
        for i, (u, s) in enumerate(zip(unique, counts)):
            if u == -1:
                continue
            centroids[0, :, i] = embedding[:, labels == u].mean(axis=-1)
            scores[0, 0, i] = s

        nonzero = torch.logical_not(centroids[0, [0,1], :].le(10).sum(0).gt(0).bool())
        centroids = centroids[:, :, nonzero]
        scores = scores[..., nonzero]

        nonzero = torch.logical_not(centroids[0, [2], :].le(1).sum(0).gt(0).bool())
        centroids = centroids[:, :, nonzero]
        scores = scores[..., nonzero]

        # assert centroids.min() >= 1, f'{nonzero}, {centroids}'

        if centroids.numel() == 0:  # nothing detected
            return torch.empty((1, 0, 3), device=device)

        # Works best with non maximum supression

        centroids_xy = centroids[0, [0, 1], :].T
        wh = torch.ones(centroids_xy.shape,
                        device=device) * 45  # <- I dont know why this works but it does so deal with it????
        boxes = torch.cat((centroids_xy, wh), dim=-1)
        boxes = torchvision.ops.box_convert(boxes, 'cxcywh', 'xyxy')
        if boxes.ndim > 2: boxes = boxes.squeeze(0)
        if boxes.shape[0] > 1:
            keep = torchvision.ops.nms(boxes.squeeze(0), scores.squeeze().float(), 0.5)
        else:
            keep = torch.ones(centroids.shape[-1]).gt(0)

        return centroids[:, :, keep].transpose(1, 2)


@torch.jit.script
def _iou(a: Tensor, b: Tensor) -> Tensor:
    """
    Takes two identically sized tensors and computes IOU of the Masks
    Assume a, b are boolean tensors with values 0 and 1

    :param a: Tensor - Bool
    :param b: Tensor - Bool

    :return: IoU
    """
    if a.dtype != torch.float: raise ValueError('input tensors must be of type float')
    val = (a + b)
    intersection = val.gt(1).sum()
    union = val.gt(0).sum()
    return intersection / union


class nms(nn.Module):
    def __init__(self):
        super(nms, self).__init__()

    def forward(self, mask: Tensor, threshold: Optional[float] = 0.5) -> Tensor:
        """
        Performs non-maximum supression on object probability masks.

        Example:

        >>> from hcat.lib.functional import nms
        >>> import torch
        >>> cell_mask = torch.load('cell_segmentation_masks.trch') #  [B, N, X, Y, Z]
        >>> nms = torch.jit.script(nms) # [1, 20, 100, 100, 10]
        >>> keep = nms(cell_mask)
        >>> cell_mask = cell_mask[:, keep, ...] # [1, 15, 100, 100, 10]

        :param mask: [B, N, X, Y, Z] predicted cell instance segmentation mask
        :param threshold: iou rejection threshold
        :return: [N] index of remaining cells
        """
        n = mask.shape[1]
        if n == 0:
            return torch.zeros(n).gt(0)

        if mask.dtype != torch.float:
            raise ValueError('Mask dtype must be float', mask.dtype)

        max_instances = mask.shape[1]
        ind = torch.ones(max_instances, dtype=torch.int, device=mask.device)
        score = mask.squeeze(0).reshape(max_instances, -1).sum(-1)

        for i in range(max_instances):
            if ind[i] == 0:
                continue
            for j in range(max_instances):
                iou = _iou(mask[:, i, ...].float(), mask[:, j, ...].float())
                if iou > threshold:
                    if score[j] > score[i]:
                        ind[i] = 0
                    if score[i] > score[j]:
                        ind[j] = 0
                        break  # break out of loop because base is now zero!

        return ind > 0


# @graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not calculate cochlear length.' + '\x1b[0m')
class PredictCurvature:
    def __init__(self,
                 voxel_dim=(.28888, .28888, .28888 * 3), # um
                 equal_spaced_distance: float = 0.01,
                 erode: int = 3,
                 scale_factor=10,
                 method=None):

        self.equal_spaced_distance = equal_spaced_distance
        self.erode = erode
        self.method = method
        self.voxel_dim = voxel_dim
        self.scale_factor = scale_factor

        _methods = ['mask', 'maxproject', None]
        if not self.method in _methods:
            raise RuntimeError(f'Unknown method: {method}.')

    def __call__(self,
                 image: Optional[Tensor] = None,
                 cells: Optional[List[Cell]] = None,
                 csv: Optional[str] = None) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Try's various methods to predict curvature.
        If user provides a csv of curvature default to this.
        In order -> CSV, Myo7a, Mask

        :param image:
        :param mask:
        :param csv:
        :return:
        """
        curvature, distance, apex = None, None, None
        if self.method is None:
            if csv is not None:
                curvature, distance, apex = self.fitEPL(csv)

            if curvature is None and cells is not None:
                try:
                    # Estimate cochlear curvature. Try the
                    _, x, y, _ = image.shape
                    mask = torch.zeros((1, x, y, 1))
                    for c in cells:
                        x0, y0, x1, y1 = torch.round(c.boxes).int()
                        mask[:, y0:y1, x0:x1, :] = 1
                    curvature, distance, apex = self.fit(mask, 'mask', diagnostic=False)

                except Exception:
                    curvature, distance, apex = None, None, None

            try_again = curvature is None and image is not None
            try_again = try_again or distance.max() < 4000

            if try_again:
                try:
                    curvature, distance, apex = self.fit(image, 'maxproject')
                except Exception:
                    curvature, distance, apex = None, None, None

        elif self.method == 'mask':
            _, x, y, _ = image.shape
            mask = torch.zeros((1, x, y, 1))
            for c in cells:
                x0, y0, x1, y1 = torch.round(c.boxes).int()
                mask[:, y0:y1, x0:x1, :] = 1
            curvature, distance, apex = self.fit(mask, 'mask', diagnostic=False)

        elif self.method == 'maxproject':
            curvature, distance, apex = self.fit(image, 'maxproject', diagnostic=False)

        else:
            raise ValueError(f'Unknown method: {self.method}')

        return curvature, distance, apex

    def fit(self, base: Tensor, method: str, diagnostic=False) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Predicts cochlear curvature from a predicted segmentation mask.

        Uses beta spline fits as well as a gaussian process regression to estimate a contiguous curve in spherical space.

        .. warning:
           Will not raise an error upon failure, instead returns None and prints to standard out

        :parameter image: [C, X, Y, Z] bool tensor
        :return: Tuple[Tensor, Tensor, Tensor]:
            - equal_spaced_points: [2, N] Tensor of pixel locations
            - percentage: [N] Tensor of cochlear percentage
            - apex: Tensor location of apex (guess)
        """

        if method == 'mask':
            image = base[0, ..., 0].bool().int()
            image = self._preprocess(image, diagnostic)
        elif method == 'stack':
            image = base[0, ...].sum(-1).gt(1.5)
            image = self._preprocess(image, diagnostic)
        elif method == 'maxproject':
            gt = 0.55
            max = 0
            while max == 0:
                image = base[0, ..., 0].gt(gt)
                image = self._preprocess(image, diagnostic)
                max = image.max()
                gt -= 0.15

        # Sometimes there are NaN or inf we have to take care of
        image[np.isnan(image)] = 0
        try:
            center_of_mass = np.array(scipy.ndimage.center_of_mass(image))
            while image[int(center_of_mass[0]), int(center_of_mass[1])] > 0:
                center_of_mass += 1
        except ValueError:
            center_of_mass = [image.shape[0], image.shape[1]]

        # Turn the binary image into a list of points for each pixel that isnt black
        x, y = image.nonzero()
        x += -int(center_of_mass[0])
        y += -int(center_of_mass[1])

        # Transform into spherical space
        r = np.sqrt(x ** 2 + y ** 2)
        theta = np.arctan2(x, y)

        # sort by theta
        ind = theta.argsort()
        theta = theta[ind]
        r = r[ind]

        # there will be a break somewhere because the cochlea isnt a full circle
        # Find the break and subtract 2pi to make the fun continuous
        loc = np.abs(theta[0:-2:1] - theta[1:-1:1])


        # Correct if largest gap is larger than 20 degrees
        if np.abs(loc.max()) > 20 / 180 * np.pi:
            theta[loc.argmax()::] += -2 * np.pi
            ind = theta.argsort()[1:-1:1]
            theta = theta[ind]
            r = r[ind]

        # run a spline in spherical space after sorting to get a best approximated fit
        tck, u = splprep([theta, r], w=np.ones(len(r)) / len(r), s=0.5e-6, k=3)

        # Run a Gaussian Process Fit, seems to work better than spline
        kernel = GPy.kern.RBF(input_dim=1, variance=50., lengthscale=5.)  # 100 before

        # Might fail, throw an error
        if np.any(np.isnan(theta)) or np.any(np.isnan(r)) or np.any(np.isinf(theta)) or np.any(np.isinf(r)):
            print('\x1b[1;31;40m' + 'ERROR: Could not calculate cochlear length.' + '\x1b[0m')
            return None, None, None

        m = GPy.models.GPRegression(theta[::2, np.newaxis], r[::2, np.newaxis], kernel)
        m.optimize()

        theta = np.linspace(theta.min(), theta.max(), 10000)
        r_, _ = m.predict(theta[:, np.newaxis])
        r_ = r_[:, 0]
        theta_ = theta

        x_spline = r_ * np.cos(theta_) + center_of_mass[1]
        y_spline = r_ * np.sin(theta_) + center_of_mass[0]

        # x_spline and y_spline have tons and tons of data points.
        # We want equally spaced points corresponding to a certain distance along the cochlea
        # i.e. we want a point ever mm which is not guaranteed by x_spline and y_spline

        # THIS ONLY REMOVES STUFF, DOESNT INTERPOLATE!!!!!!
        curvature = []
        for i, coord in enumerate(zip(x_spline, y_spline)):
            if i == 0:
                base = coord
                curvature.append(base)
            if np.sqrt((base[0] - coord[0]) ** 2 + (base[1] - coord[1]) ** 2) > self.equal_spaced_distance:
                curvature.append(coord)
                base = coord

        curvature = np.array(curvature) * self.scale_factor # <-- Scale factor from above
        curvature = curvature.T

        # Figure out the knot with the larget value and which side its closest to. Thats probably the apex
        # If the curviest point is on the first half of the curve or second half. 1
        curve = tck[1][0]
        curviest_point = np.abs(curve).argmax()
        # curviest_point_a = np.abs(curve[:curve.shape[0]//3:]).argmax()
        # curviest_point_b = np.abs(curve[2*curve.shape[0]//3::]).argmax()
        #
        apex = curvature[:, -1] if curviest_point < curve.shape[0] // 2 else curvature[:, 0]
        # apex = curvature[:, -1] if curviest_point_b > curviest_point_a else curvature[:, 0]

        # sort curvature so that it ALWAYS GOES base to apex
        # curvature = curvature[:, ::-1] if curviest_point_b > curviest_point_a else curvature
        curvature = curvature[:, ::-1] if np.all(apex == curvature[:, -1]) else curvature

        # Calculate the distance from the apex
        distance = np.zeros(curvature.shape[1])
        x0, y0 = curvature[:, 0]
        for i in range(1, curvature.shape[1]):
            x, y = curvature[:, i]
            dx = ((x - x0) * self.voxel_dim[0]) ** 2
            dy = ((y - y0) * self.voxel_dim[1]) ** 2
            distance[i] = np.sqrt(dx + dy) + distance[i-1]
            x0, y0 = curvature[:, i]

        return torch.from_numpy(curvature.copy()), torch.from_numpy(distance.copy()), torch.from_numpy(apex.copy())

    def fitEPL(self, path: str, pix2um: float = 3.4616):
        """

        :param path:
        :return:
        """

        # csv = np.genfromtxt(path, delimiter=',', skip_header=1)
        csv = pandas.read_csv(path)

        x: Tensor = torch.tensor(csv['X'].to_list())
        y: Tensor = torch.tensor(csv['Y'].to_list())
        center = torch.cat([x.unsqueeze(0), y.unsqueeze(0)]).mean(dim=1)

        x = x - center[0]
        y = y - center[1]

        theta = torch.arctan(y/x)
        r = torch.sqrt(x**2 + y**2)

        loc = torch.abs(theta[0:-2:1] - theta[1:-1:1])

        # if np.abs(loc.max()) > 20 / 180 * np.pi:
        #     theta[loc.argmax()+1::] += 2*np.pi

        tck, u = splprep([theta.numpy(), r.numpy()])#, w=np.ones(len(r)) / len(r), s=1e-2, k=3)
        u_new = np.linspace(0, 1, 1000)
        theta_, r_ = splev(u_new, tck)
        theta_ = torch.from_numpy(theta_)
        r_ = torch.from_numpy(r_)

        x_spline = r_ * torch.cos(theta_) + center[0]
        y_spline = r_ * torch.sin(theta_) + center[1]

        tck, u = splprep([x.numpy(), y.numpy()])#, w=np.ones(len(r)) / len(r), s=1e-2, k=3)
        u_new = np.linspace(0, 1, 1000)
        x_, y_ = splev(u_new, tck)

        x_ = torch.from_numpy(x_) + center[0]
        y_ = torch.from_numpy(y_) + center[1]
        equal_spaced_distance = 0.0001
        curvature  = []
        for i, coord in enumerate(zip(x_, y_)):
            if i == 0:
                base = coord
                curvature .append(base)
            if np.sqrt((base[0] - coord[0]) ** 2 + (base[1] - coord[1]) ** 2) > equal_spaced_distance:
                curvature.append(coord)
                base = coord

        curvature = torch.tensor(curvature).mul(pix2um).T  # <-- mm to px
        # curvature = curvature.T

        # Assume manual is always base to apex
        apex = curvature[:, -1]

        # sort curvature so that it ALWAYS GOES base to apex - UNNESSECARY!
        # curvature = curvature[:, ::-1] if np.all(apex == curvature[:, 0]) else curvature

        total_distance = 0
        distance = torch.zeros(curvature.shape[1])
        x0, y0 = curvature[:, 0]
        # calculate the distance from
        for i in range(1, curvature.shape[1]):
            # each pixel is 288.88nm
            x, y = curvature[:, i]
            dx = ((x - x0) * self.voxel_dim[0]) ** 2
            dy = ((y - y0) * self.voxel_dim[1]) ** 2
            distance[i] = np.sqrt(dx + dy) + distance[i-1]
            x0, y0 = curvature[:, i]

        return curvature, distance, apex

    def _preprocess(self, image, diagnostic):

        image = skimage.transform.downscale_local_mean(image.numpy(), (10, 10)) > 0
        image = skimage.morphology.binary_closing(image)
        image = skimage.morphology.diameter_closing(image, 10)

        if diagnostic:
            plt.imshow(image)
            plt.show()

        for i in range(self.erode):
            image = skimage.morphology.binary_erosion(image)
        for i in range(self.erode):
            image = skimage.morphology.binary_dilation(image)

        if diagnostic:
            plt.imshow(image)
            plt.show()

        return skimage.morphology.skeletonize(image, method='lee')



########################################################################################################################
#                                                 U Net
########################################################################################################################

class PredictCellCandidates(nn.Module):
    def __init__(self, model: nn.Module, device: str = 'cpu'):
        """
        Initializes a Faster RCNN model for detection of cell candidates.
        These cell candidates are used as seed values for a watershed segmentation algorithm.

        :param model: Initialized faster RCNN model
        :param device: device used to perform calculations ('cpu' or 'cuda')
        """
        super(PredictCellCandidates, self).__init__()
        self.model = model.to(device)
        self.device = device

    def forward(self, image: Tensor) -> Dict[str, Tensor]:
        """
        Applies a FasterRCNN model on each z plane of an image. Returns the prediction in a dictionary.

        Example:

        >>> from hcat.lib.functional import PredictCellCandidates
        >>> import torch
        >>> image = torch.load('image.trch') # [B, C, X, Y, Z] ground truth mask
        >>> predict = PredictCellCandidates()
        >>> centroids: Dict[str, Tensor] = predict(image) # keys: 'scores', 'boxes', 'labels'

        :param image: [B, C=4, X, Y, Z] input image
        :return: a dictionary with keys: 'scores', 'boxes', 'labels'.
        """

        # Check Inputs
        if not isinstance(image, Tensor): raise ValueError(f'Image should be torch.tensor {type(image)}')
        if image.ndim != 5: raise ValueError(f'Image dimmensions should be 5, not {image.ndim}')
        if image.shape[0] != 1: raise ValueError(
            f'Multiple Batches not supported. Image.shape[0] should be 1, not {image.shape[0]}')

        max_num_cells = 0
        out = {'boxes': torch.tensor([]), 'scores': torch.tensor([]), 'labels': torch.tensor([])}

        with torch.no_grad():
            for z in range(image.shape[-1]):  # ASSUME TORCH VALID IMAGES [B, C, X, Y, Z]

                # Take a mini chunk out of the original image
                image_z_slice = image[..., z]

                # Apply faster rcnn prediction model to this
                # We're only doing a batch size of 1, take the first dict of results
                # list of dicts (batch size) with each dict containing 'boxes' 'labels' 'scores'
                predicted_cell_locations = self.model(image_z_slice.to(self.device))[0]

                # We need to know what z plane these cells were predicted, add this as a new index to the dict
                z_level = torch.ones(len(predicted_cell_locations['scores'])) * z
                predicted_cell_locations['z_level'] = z_level

                if len(predicted_cell_locations['scores']) > len(out['scores']):
                    out = predicted_cell_locations

            # Move everything to be a float and on the cpu()
            for name in out:
                out[name] = out[name].to(self.device).float()

        return out


class PredictSemanticMask(nn.Module):
    def __init__(self,
                 model: nn.Module,
                 device: str = 'cpu',
                 use_probability_map: bool = False,
                 mask_cell_prob_threshold: float = 0.5):
        super(PredictSemanticMask, self).__init__()

        self.device = device
        self.use_prob_map = use_probability_map
        self.mask_threshold = mask_cell_prob_threshold

        self.unet = model.to(device)

    def forward(self, image: Tensor) -> Tensor:
        """
        Uses pretrained unet model to predict semantic segmentation of hair cells.

        ALGORITHM:
        Remove inf and nan ->
        apply padding ->
        calculate indexes for unet based on image.shape ->
        apply Unet on slices of image based on indexes ->
        take only valid portion of the middle of each output mask ->
        construct full valid mask ->
        RETURN mask


        BLUR PROBABILITY MAP???

        :param unet: Trained Unet Model from hcat.unet
        :param image: Tensor image with transforms pre applied!!! [1, 4, X, Y, Z]
        :param device: 'cuda' or 'cpu'
        :param mask_cell_prob_threshold: float between 0 and 1 : min probability of cell used to generate the mask
        :param use_probability_map: bool, if True, does not apply sigmoid function to valid out, returns prob map instead
        :return mask:

        """

        # Check Inputs
        if not isinstance(image, Tensor): raise ValueError(f'Image should be torch.tensor {type(image)}')
        if image.ndim != 5: raise ValueError(f'Image dimmensions should be 5, not {image.ndim}')
        if image.shape[0] != 1: raise ValueError(
            f'Multiple Batches not supported. Image.shape[0] should be 1, not {image.shape[0]}')

        PAD_SIZE = (30, 30, 4)

        # inf and nan screw up model evaluation. Happens occasionally. Remove them!
        image[torch.isnan(image)] = 0
        image[torch.isinf(image)] = 1

        im_shape = image.shape

        # Apply Padding by reflection to all edges of image
        # With pad size of (30, 30, 4)
        # im.shape = [1,4,100,100,10] -> [1, 4, 160, 160, 18]
        image = hcat.lib.utils.pad_image_with_reflections(torch.as_tensor(image), pad_size=PAD_SIZE)

        with torch.no_grad():

            valid_out = self.unet(image.to(self.device))

            # Unet cuts off bit of the image. It is unavoidable so we add padding to either side
            # We need to remove padding from the output of unet to get a "Valid" segmentation
            valid_out = valid_out[:, :,
                        PAD_SIZE[0]:im_shape[2] + PAD_SIZE[0],
                        PAD_SIZE[1]:im_shape[3] + PAD_SIZE[1],
                        PAD_SIZE[2]:im_shape[4] + PAD_SIZE[2]]

            # Take pixels that are greater than 50% likely to be a cell.
            if not self.use_prob_map:
                valid_out.gt_(self.mask_threshold)  # Greater Than
                valid_out = valid_out.type(torch.uint8)

        return valid_out


class GenerateSeedMap(nn.Module):
    def __init__(self,
                 cell_prob_threshold: float = 0.5,
                 mask_prob_threshold: float = 0.5):
        super(GenerateSeedMap, self).__init__()

        self.cell_threshold = cell_prob_threshold
        self.mask_threshold = mask_prob_threshold

    def forward(self, cell_candidates: Dict[str, Tensor], prob_map: Tensor) -> Tensor:
        """

        :param cell_candidates: Dict[str, Tensor] keys = ['scores', 'boxes', 'z_plane', 'labels']
        :param prob_map: [B, C, X, Y, Z] Tensor
        :return:
        """

        unique_cell_id = 2  # 1 is reserved for background
        expand_z = 4  # dialate the z to try and account for nonisotropic zstacks
        b, c, x, y, z = prob_map.shape
        seed = torch.zeros(prob_map.shape, dtype=np.int)

        if prob_map.max() == 0 or len(cell_candidates['scores']) == 0:
            return seed

        z = cell_candidates['z_level'].cpu()
        scores = cell_candidates['scores'].cpu()
        boxes = cell_candidates['boxes']

        z = z[scores > self.cell_threshold]
        prob = scores[scores > self.cell_threshold]

        if prob.sum() == 0:
            return torch.zeros(prob_map.shape)

        boxes = boxes[scores > self.cell_threshold, ...]

        # Try to basically remove boxes that arent likely to be over a cell.
        x1, y1, x2, y2 = boxes.cpu().T
        centers = torch.cat((torch.round(x1 + (x2 - x1)).unsqueeze(0),
                             torch.round(y1 + (y2 - y1)).unsqueeze(0),
                             z.unsqueeze(0)
                             ), dim=0)

        # Loop over every center and see if it lies over a region of high probability on the probability map
        # This prunes cell candidates to reasonable ones
        ind = torch.zeros(len(x1))
        for i, (x0, y0, z0) in enumerate(centers.T):
            try:
                if prob_map[0, 0, int(x0), int(y0), int(z0)] > 0.5:
                    ind[i] = 1
            except Exception:
                pass

        # We look at the cell candidates z planes and choose just one that is optimal
        # This is done by looking at the larges z plane with the biggest mean score.
        unique_z, counts = torch.unique(z, return_counts=True)
        best_z = 0
        best_z_avg = 0
        for uni in unique_z:
            if prob[z == uni].mean() > best_z_avg:
                best_z = uni
                best_z_avg = prob[z == uni].mean()

        # Sort to put close centers near each other
        sorted, indices = torch.sort(boxes, dim=0)
        for key in cell_candidates:
            cell_candidates[key] = cell_candidates[key][indices[:, 0], ...]

        # Generate Seed Map
        for i, (y1, x1, y2, x2) in enumerate(cell_candidates['boxes']):

            # There are various conditions where a box is invalid
            # in these cases, do not place a seed and skip the box
            if x1 > prob_map.shape[2]:
                continue  # box is outside x dim
            elif y1 > prob_map.shape[3]:
                continue  # box is outside y dim
            elif cell_candidates['scores'][i] < self.cell_threshold:
                continue  # box probability is lower than predefined threshold
            elif cell_candidates['z_level'][i] < best_z - 2:
                continue  # box is on the wrong z plane within a tolerance
            elif cell_candidates['z_level'][i] > best_z + 2:
                continue  # box is on the wrong z plane within a tolerance

            # in the cases where a box is clipping the outside, crop it to the edges
            if x2 > prob_map.shape[2]:
                x2 = torch.tensor(prob_map.shape[2] - 1).float()
            elif y2 > prob_map.shape[3]:
                y2 = torch.tensor(prob_map.shape[3] - 1).float()

            # Each box is a little to conservative in its estimation of a hair cell
            # To compensate, we add dx and dy to the corners to increase the size
            dx = [5, -5]
            dy = [5, -5]

            # Check if adding will push box outside of range!
            if (x1 + dx[0]) < 0:
                dx[0] = x1
            if (y1 + dy[0]) < 0:
                dy[0] = y1
            if (x2 + dx[1]) > prob_map.shape[2]:
                dx[1] = prob_map.shape[2] - x2
            if (y2 + dy[1]) > prob_map.shape[3]:
                dy[1] = prob_map.shape[3] - y2

            x1 = int(x1.clone().add(dx[0]).round())
            x2 = int(x2.clone().add(dx[1]).round())
            y1 = int(y1.clone().add(dy[0]).round())
            y2 = int(y2.clone().add(dy[1]).round())

            box = prob_map[0, 0, int(x1):int(x2), int(y1):int(y2), int(best_z)]

            # Here we place a seed value for watershed at each point of the valid box
            seed_square = 2
            for i in range(seed_square):
                if i + best_z > seed.shape[-1]:
                    continue
                try:
                    seed[0, 0, int(x1):int(x2), int(y1):int(y2), int(best_z) + i][box == box.max()] = int(
                        unique_cell_id)
                except IndexError:
                    if seed.max() > 0:
                        continue
                    else:
                        raise IndexError(f'No Seed was placed when index')
                except ValueError:
                    continue
            unique_cell_id += 1

        return seed


class InstanceMaskFromProb(nn.Module):
    def __init__(self,
                 mask_prob_threshold: float = 0.5,
                 expand_z: int = 5,
                 use_prob_map: bool = False):
        super(InstanceMaskFromProb, self).__init__()

        self.mask_threshold = mask_prob_threshold
        self.use_prob_map = use_prob_map
        self.expand_z = expand_z

    def forward(self, semantic_mask: Tensor,
                seed: Tensor) -> Tensor:
        """

        :param semantic_mask:
        :param seed:
        :return:
        """

        if seed.max() == 0:
            b, c, x, y, z = seed.shape
            return torch.zeros((b, 0, x, y, z))

        assert semantic_mask.shape == seed.shape

        # Preallocate some matrices
        mask = semantic_mask.gt(self.mask_threshold)

        # Context aware probability map
        # If true, watershed is done on the probability map, distance is now probability
        # If false, run a distance transform on the binary mask and run watershed on that
        base = self._prepare_watershed_base(semantic_mask)

        base = self._expand_z_dimmension(base, self.expand_z)
        seed = self._expand_z_dimmension(seed, self.expand_z)
        mask = self._expand_z_dimmension(mask, self.expand_z)

        assert mask.shape == seed.shape
        assert base.shape == mask.shape

        # Run the watershed algorithm
        mask = skimage.segmentation.watershed(image=base.cpu().numpy() * -1,
                                              markers=seed.cpu().numpy(),
                                              mask=mask.cpu().numpy(),
                                              connectivity=1,
                                              watershed_line=True, compactness=0.01)

        mask = torch.from_numpy(mask)
        mask = self._correct_predicted_instance_mask(mask)

        return mask.unsqueeze(0).unsqueeze(0)

    def _prepare_watershed_base(self, semantic_mask: Tensor):
        """
        Prepare the base matrix which watershed is run off of. Has two potential methods depending on context.
            1: use distance transform of semantic segmentation mask
            2: use the probability map predicted by the algorithm

        :param semantic_mask:
        :return:
        """
        distance = torch.zeros(semantic_mask.shape)
        if self.use_prob_map and semantic_mask.max() > 1:
            semantic_mask.add_(1e-8).sub_(torch.min(semantic_mask)).div_(torch.max(semantic_mask))
            distance[0, 0, :, :, :] = semantic_mask[0, 0, :, :, :]
        else:
            for z in range(distance.shape[-1]):
                distance[0, 0, :, :, z] = torch.from_numpy(
                    scipy.ndimage.morphology.distance_transform_edt(
                        semantic_mask[0, 0, :, :, z].cpu().numpy().astype(np.uint8)))
        return distance

    @staticmethod
    def _expand_z_dimmension(matrix, expand_z):
        """
        Accounts for anisotropy by expanding each z plane of a matrix with copies
        example
        matrix: Tensor = torch.rand([1, 1, 100, 100, 10])
        matrix: Tensor = _expand_z_dimmension(matrix, expand_z = 4)
        matrix.shape
            >>> torch.Shape([1, 100, 100, 40])
        torch.all(matrix[0,:,:,1] == matrix[0,:,:,2])
            >>> True


        :param matrix: Tensor with 5 dimmensions
        :param expand_z: integer multiple
        :return:
        """
        if matrix.ndim != 5: raise ValueError(f'Expected matrix.ndim == 5 matrix not {matrix.ndim}.')
        mat_expanded = torch.zeros((matrix.shape[2], matrix.shape[3], matrix.shape[4] * expand_z))
        for i in range(matrix.shape[4]):
            for j in range(expand_z):
                mat_expanded[:, :, (expand_z * i + j)] = matrix[0, 0, :, :, i]
        return mat_expanded

    def _correct_predicted_instance_mask(self, mask: Tensor) -> Tensor:
        mask[mask == 1] = 0
        mask = mask[..., ::self.expand_z]
        return mask


########################################################################################################################
#                                               Cell Rejection
########################################################################################################################

class IntensityCellReject(nn.Module):
    def __init__(self, threshold: float = 0.08):
        """
        :param threshold: If cell avg val is lower than this, get rid of it...
        """
        super(IntensityCellReject, self).__init__()
        self.threshold = threshold

    def forward(self, mask: Tensor, image: Tensor) -> Tensor:
        """
        Checks the type of mask: colormask or binary mask.
        :param mask:
        :param image:
        :return:
        """
        mask, image = hcat.lib.utils.crop_to_identical_size(mask, image)

        # strange edge case where

        if mask.shape[1] == 1 and mask.max() >= 1:
            return self.color(mask, image)
        elif mask.shape[1] >= 1 and mask.max() < 1:
            return self.binary(mask, image)
        elif mask.shape[1] == 0:
            return torch.empty((0)).gt(0)
        else:
            raise RuntimeError(f'Unknown Mask Shape: {mask.shape}, {mask.max()}')

    def color(self, mask: Tensor, image: Tensor) -> Tensor:
        ind = torch.ones(mask.shape[1], dtype=torch.long, device=mask.device)

        unique = torch.unique(mask)
        for u in unique:
            if u == 0:
                continue
            m = (mask == u)

            if (m * image).sum().div(m.sum()) < ((self.threshold - 0.5) / 0.5):
                mask[m] = 0

        return mask

    def binary(self, mask: Tensor, image: Tensor) -> Tensor:
        """
        Re
        :param mask: [B, N, X, Y, Z]
        :param image: [B, 1, X, Y, Z]
        :return:
        """

        mask, image = hcat.lib.utils.crop_to_identical_size(mask, image)
        ind = torch.ones(mask.shape[1], dtype=torch.long, device=mask.device)

        for i in range(mask.shape[1]):
            m = mask[:, i, ...].unsqueeze(1).gt(0.5)
            if (m * image).sum().div(m.sum()) < ((self.threshold - 0.5) / 0.5):
                ind[i] = 0

        return mask[:, ind.gt(0), ...]


# @graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not merge regions.' + '\x1b[0m')
# @torch.jit.script
def merge_regions(destination: Tensor,
                  data: Tensor,
                  threshold: float = 0.25) -> Tuple[Tensor, Tensor]:
    """
    assume [C, X, Y, Z] shape for all tensors

    Takes data and puts it in destination intellegently.
    :param destination:
    :param data:
    :return:
    """
    expanded = torch.zeros(destination.shape, dtype=destination.dtype, device=data.device)

    # x_pad = (destination.shape[1] - data.shape[1]) // 2
    # y_pad = (destination.shape[2] - data.shape[2]) // 2

    expanded[:, 0:data.shape[1], 0:data.shape[2], 0:data.shape[3]] = data

    # Overlapping pixels will have a value not identical to either when summed
    overlap = destination + expanded
    overlap = torch.logical_not(torch.logical_and(overlap == expanded,
                                                  overlap == destination))

    unique_destination, counts_destination = torch.unique(destination[overlap], return_counts=True)
    unique_data, counts_data = torch.unique(expanded[overlap], return_counts=True)

    # These probably will be small so wont worry about nested loops
    for u_des, c_des in zip(unique_destination, counts_destination):
        if u_des == 0:
            continue
        a = destination == u_des
        for u_data, c_data in zip(unique_data, counts_data):
            if u_data == 0:
                continue
            b = expanded == u_data

            # If iou of each thing is greater than threshold, use branchless programing
            # To make things zero or not
            if _iou(a.float(), b.float()) > threshold:
                destination[a] *= (c_des > c_data)
                expanded[b] *= (c_des < c_data)

    destination[expanded != 0] = expanded[expanded != 0]
    return destination, torch.unique(expanded, return_counts=False)


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

# Modified by Chris Buswinka to somehow work with 4 Channel MEEI Lif files


#  The code was origionally done by Mathieu Leocmach
#  It was part of the Colloids library (https://github.com/MathieuLeocmach/colloids)
#  It is modified by Yushi Yang (yushi.yang@bristol.ac.uk) to be more compatable
#  with different python version and confocal machines
#  Since its' original version use GPL License, it's restricted GPL as well


#    Modification June 2019 by M. Leocmach /C. Ybert:
#
#        0- Note about good practices: semi-private attributes should only be accessed from a method getXXX(). Semi-privacy is chosen here to emphasize
#           that values are not supposed to be changed (they correspond to the *fixed* experimental conditions)
#
#        1- Limit read_lif to its main purpose (remove analysis functions getNeighbourhood() and getRadius, and associated library import)
#        2- Add a getFilterSetting method to SerieHeader Class
#        3- Modify Header.parse method (remove initial storage of all timestamps)
#        5- Expand (a bit) the function description
#        6- Move all double underscore attributes to simple underscore (semi-private ones) and corrected uneffective hasattr() tests
#        7- Remove get2DImage() methods
#        8- Add a getFrame2D() method to extract a 2D image. As for getFrame() it lacks proper conditional testing to accomodate (or report error)
#           depending on the type of data (XYT, XYZT, ...). It is based on the inner loop in getFrame so that it might be possible to
#           rewrite getFrame() using getFrame2D().
#
#
#        BUGS: - getMetadata() doesn't support XY frames (no z)
#              - getFrame() doesn't support XY frames (no z)
#              - get2DStrings() doesn't support XY frames (no z)
#              - enumByFrame() and enumBySlice() lacks some conditional testing against the data frame type
#
#
#        ToDo: - Correct Bugs
#              - Convert TimeStamps into seconds
#              - getZXratio it would be more consistent to use getDimensions() to obtain the information
#              - Automatic type conversion of SeriesHeader data (in ScannerSetting and FilterSetting) using the VariantType data from VisualBasic
#                {0: Empty, 1: Null, 2: Short, 3: integer, 4: single, 5: double, 6: currency, 7: date, 8: string, 9: object, 10: error
#                 11: boolean, 12: variant, 13: dataObject, 14: decimal, 17: byte, 18: char, 20: long, 36: UserDefined, 8192: array}


# ============== origional information ====================
#
#    Copyright 2009 Mathieu Leocmach
#
#    This file is part of Colloids.
#
#    Colloids is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Colloids is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Colloids.  If not, see <http://www.gnu.org/licenses/>.
#

import struct, io, re, sys
import xml
from xml.dom.minidom import parse
import numpy as np
import warnings

dimName = {1: "X",
           2: "Y",
           3: "Z",
           4: "T",
           5: "Lambda",
           6: "Rotation",
           7: "XT Slices",
           8: "TSlices",
           10: "Tiles"}

channelTag = ["Blue", "Green", "Yellow", "Red"]


class Header:
    """
    The XML header of a Leica LIF files

    Attributes:: are all semi-private (should only be accessed from a method getXXX())
                 as values are not supposed to be changed (they correspond to the *fixed* experimental conditions)
                 _version, _name, _seriesHeaders,
    """

    def __init__(self, xmlHeaderFileName, quick=True):
        if sys.version_info > (3, 0):
            with open(xmlHeaderFileName, encoding="latin-1") as f:
                self.parse(f, quick)
        else:
            with open(xmlHeaderFileName) as f:
                self.parse(f, quick)

    def parse(self, xmlHeaderFile, quick=False):
        """
        Method: Parse the usefull part of the xml header. Non ascii characters are stripped
        Args:

            quick (Boolean). If True Time Stamps are also stripped from the xmlHeader

        .. warnings:: `quick == True` forbids future access to TimeStamps unless re-opening of original .lif file
        """

        # to strip the non ascii characters
        t = "".join(map(chr, list(range(256))))
        d = "".join(map(chr, list(range(128, 256))))
        if sys.version_info > (3, 0):
            trans = str.maketrans('', '', d)
            lightXML = io.StringIO()
        else:
            import StringIO
            trans = {ord(c): None for c in d}
            lightXML = StringIO.StringIO()

        if not quick:
            xmlHeaderFile.seek(0)
            if sys.version_info > (3, 0):
                for line in xmlHeaderFile:
                    lightXML.write(line.translate(trans))
            else:
                for line in xmlHeaderFile:
                    try:
                        lightXML.write(line.translate(t, d))
                    except TypeError:
                        lightXML.write(line.translate(trans))

        else:
            # to strip the time stamps
            m = re.compile(
                r'''<TimeStamp HighInteger="[0-9]*" LowInteger="[0-9]*"/>|'''
                + r'''<RelTimeStamp Time="[0-9.]*" Frame="[0-9]*"/>|'''
                + r'''<RelTimeStamp Frame="[0-9]*" Time="[0-9.]*"/>'''
            )
            if sys.version_info > (3, 0):
                for line in xmlHeaderFile:
                    lightXML.write(''.join(m.split(line)).translate(trans))
            else:
                for line in xmlHeaderFile:
                    try:
                        lightXML.write(''.join(m.split(line)).translate(t, d))
                    except TypeError:
                        lightXML.write(''.join(m.split(line)).translate(trans))
        lightXML.seek(0)
        self.xmlHeader = parse(lightXML)

    def getVersion(self):
        """
        Method: Get the version of the Data Container Header
        Semi-private attribute `_version`
        """
        if not hasattr(self, '_version'):
            self._version = self.xmlHeader.documentElement.getAttribute("Version")
        return int(self._version)

    def getName(self):
        """
        Method: Get the name of current lif file (without extension .lif)
        Semi-private attribute `_name`
        """
        if not hasattr(self, '_name'):
            self._name = self.xmlHeader.documentElement. \
                getElementsByTagName('Element')[0].getAttribute("Name")
        return self._name

    def getSeriesHeaders(self):
        """
        Method: Get the Series Headers of all series contained in the .lif file
        Semi-private attribute `_seriesHeaders`
        """

        if not hasattr(self, '_seriesHeaders'):
            root = self.xmlHeader.documentElement
            headers = []
            counter = 0
            for element in root.getElementsByTagName('Element'):
                element_nodes = [node for node in element.childNodes if node.ELEMENT_NODE == 1]
                memory_nodes = [node for node in element_nodes if (node.localName == 'Memory')]
                if len(memory_nodes) > 0:
                    counter += 1
                    memory_nodes = memory_nodes[0]
                    size = memory_nodes.getAttribute('Size')
                    if size:
                        if int(size) > 0:
                            headers.append(SerieHeader(element))
            self._seriesHeaders = headers
        return self._seriesHeaders

    def chooseSerieIndex(self):
        """
        Method: Interactive selection of the desired serie.

        Print a list of all Series specifying: index, Name, Channel number and Tag, and stack dimensions X, Y, Z, T,...
        Promt for selecting a Serie number.

        """
        st = "Experiment: %s\n" % self.getName()
        for i, s in enumerate(self.getSeriesHeaders()):
            # s = Serie(serie)
            st += "(%i) %s: %i channels and %i dimensions\n" % (
                i, s.getName(), len(s.getChannels()), len(s.getDimensions())
            )
            for c in s.getChannels():
                st += " %s" % channelTag[int(c.getAttribute("ChannelTag"))]
            for d in s.getDimensions():
                st += " %s%i" % (
                    dimName[int(d.getAttribute("DimID"))],
                    int(d.getAttribute("NumberOfElements"))
                )
            st += "\n"
        self.x = print(st)
        if len(self.getSeriesHeaders()) < 2:
            r = 0
        else:
            while (True):
                try:
                    r = int(input("Choose a serie --> "))
                    if r < 0 or r > len(self.getSeriesHeaders()):
                        raise ValueError()
                    break
                except ValueError:
                    print("Oops!  That was no valid number.  Try again...")
        return r

    def chooseSerieHeader(self):
        """
        Method: Interactive selection of the desired serie header.

        Print a list of all Series specifying: index, Name, Channel number and Tag, and stack dimensions X, Y, Z, T,...
        Promt for selecting a Serie number.

        """

        return self.getSeriesHeaders()[self.chooseSerieIndex()]

    def __iter__(self):
        return iter(self.getSeriesHeaders())


class SerieHeader:
    """
    The part of the XML header of a Leica LIF files concerning a given serie

    Attributes:: are all semi-private (should only be accessed from a method getXXX())
                 as values are not supposed to be changed (they correspond to the *fixed* experimental conditions)
                 _name, _isPreview, _channels, _dimensions, _memorySize, _resolution, _numberOfElements, _duration,
                 _timeStamps, _relTimeStamps, _nbFrames, _boxShape, _nbPixelsPerFrame, _nbPixelsPerSlice, etc.

    """

    def __init__(self, serieElement):
        self.root = serieElement

    def getName(self):
        """
        Method: Get the Name of the Serie
        Semi-private attribute `_name`
        """
        if not hasattr(self, '_name'):
            self._name = self.root.getAttribute("Name")
        return self._name

    def isPreview(self):
        if not hasattr(self, '_isPreview'):
            self._isPreview = 0
            for c in self.root.getElementsByTagName("Attachment"):
                if c.getAttribute("Name") == "PreviewMarker":
                    self._isPreview = bool(c.getAttribute("isPreviewImage"))
                    break
        return self._isPreview

    def getChannels(self):
        """
        Method: Extract the DOM elements describing the used channels.
        Semi-private attribute `_channels`

        Return: List of DOM elements

        Use: To recover information contained in `chosen_serie.getChannels()`, iterate over the elements and apply `.getAttribute('Keyword')` method

        Keywords: Refer to xml file for a list of relevant Keywords. Examples: DataType, ChannelTag, Resolution, NameOfMeasuredQuantity, Min, Max, Unit,
                  LUTName, IsLUTInverted, BytesInc, BitInc
        """
        if not hasattr(self, '_channels'):
            self._channels = self.root.getElementsByTagName("ChannelDescription")
        return self._channels

    def getDimensions(self):
        """
        Method: Extract the DOM elements describing the used channels.
        Semi-private attribute `_dimensions`

        Return: List of DOM elements

        Use: To recover information contained in `chosen_serie.getDimensionss()`, iterate over the elements and apply `.getAttribute('Keyword') method

        Keywords: Refer to xml file for a list of relevant Keywords. Examples: DimID, NumberOfElements, Origin, Length, Unit, BitInc, BytesInc
        """
        if not hasattr(self, '_dimensions'):
            self._dimensions = self.root.getElementsByTagName(
                "DimensionDescription")
        return self._dimensions

    def hasZ(self):
        """
        Method: Check if current serie includes Z stacking

        Return: Boolean value
        """
        for d in self.getDimensions():
            if dimName[int(d.getAttribute("DimID"))] == "Z":
                return True
        return False

    def getMemorySize(self):
        """
        Method: Get the Memory size of the current Serie
        Semi-private attribute `_memorySize`

        Return: Memory size as int
        """
        if not hasattr(self, '_memorySize'):
            for m in self.root.getElementsByTagName("Memory"):
                # to ensure the Memory node is the child of root
                if m.parentNode is self.root:
                    self._memorySize = int(m.getAttribute("Size"))
        return self._memorySize

    def getResolution(self, channel):
        """
        Method: Get the BitDepth Resolution of a given Channel using .getChannels() method
        Semi-private attribute `_resolution`

        arg:: channel number (int)

        Return: Bit Depth resolution as int
        """
        if not hasattr(self, '_resolution'):
            self._resolution = int(
                self.getChannels()[channel].getAttribute("Resolution")
            )
        return self._resolution

    def getScannerSetting(self, identifier):
        """
        Method: Access to the value of one of the Experimental Condition elements under ScannerSettingRecord TagName
        Semi-private attribute `_"identifier"`

        arg:: identifier (string) as the name of the Attribute looked for.

              Refer to xml file for a list of relevant identifier: dblPinhole, dblSizeX, dblSizeY, dblVoxelX, dblVoxelY, dblZoom,
              nAccumulation, nAverageFrame, nAverageLine, nChannels, nDelayTime_ms, nLineAccumulation, nRepeatActions, ...

        return: The value ('Variant') of the required Attribute ('identifier') as a *string*
        """
        if not hasattr(self, '_' + identifier):
            for c in self.root.getElementsByTagName("ScannerSettingRecord"):
                if c.getAttribute("Identifier") == identifier:
                    setattr(self, '_' + identifier, c.getAttribute("Variant"))
                    break
        return getattr(self, '_' + identifier)

    def getFilterSetting(self, objectName):
        """
        Method: Access to the value of one of the Experimental Condition elements under `FilterSettingRecord TagName`
        Semi-private attribute `_"objectName"`

        arg:: objectName (string) as the name of the Object looked for.

              Refer to xml file for a list of relevant objectName: e.g. 'Scan Head' yields information about scanner

        return: Unlike identifier in getScannerSetting, a single objectname has multiple occurences,
                hence getFilterSetting returns a dictionnary of all {Attributes: Variant}, with values 'Variant' being strings
        """
        if not hasattr(self, '_' + objectName):
            obj = dict()
            for c in self.root.getElementsByTagName("FilterSettingRecord"):
                if c.getAttribute("ObjectName") == objectName:
                    obj[c.getAttribute("Attribute")] = c.getAttribute("Variant")
            setattr(self,  '_' + objectName, obj)
        return getattr(self, '_' +  objectName)

    def getNumberOfElements(self):
        """
        Method: Use .getDimensions() method to extact the data dimensions along all axis in order X, Y, Z, T, ...
        Semi-private attribute `_numberOfElements`

        Result: List of integers
        """
        if not hasattr(self, '_numberOfElements'):
            self._numberOfElements = [
                int(d.getAttribute("NumberOfElements")) \
                for d in self.getDimensions()
            ]
        return self._numberOfElements

    def getVoxelSize(self, dimension):
        """
        Method: Use getScannerSetting() to return the resolution

        arg:: dimension is an integer according to the dimName dictionnary {1: X, 2: Y, 3: Z}

        Return: Voxel size of specified axis (float)

        """
        return float(self.getScannerSetting("dblVoxel%s" % dimName[dimension]))

    def getZXratio(self):
        """
        Method: Calculate for the current Serie the ratio between Vertical (Z) and Horizontal (X) image resolutions. Looks for the information either in
                    - 'ScannerSettingRecord' (resolution)
                    - or in 'DimensionDescription' (total size / pixel number)

        Return: Z/X resolution ratio (float). Return 1.0 if the current Serie has no Z-stack.

        """
        setting_records = self.root.getElementsByTagName('ScannerSettingRecord')
        dimension_descriptions = self.root.getElementsByTagName('DimensionDescription')
        if self.hasZ():
            if setting_records:
                return float(self.getScannerSetting("dblVoxelZ")) / float(self.getScannerSetting("dblVoxelX"))
            elif dimension_descriptions:
                length_x = float(
                    [d.getAttribute('Length') for d in dimension_descriptions if d.getAttribute('DimID') == '1'][0])
                length_z = float(
                    [d.getAttribute('Length') for d in dimension_descriptions if d.getAttribute('DimID') == '3'][0])
                number_x = float(
                    [d.getAttribute('NumberOfElements') for d in dimension_descriptions if
                     d.getAttribute('DimID') == '1'][
                        0])
                number_z = float(
                    [d.getAttribute('NumberOfElements') for d in dimension_descriptions if
                     d.getAttribute('DimID') == '3'][
                        0])
                psx = length_x / number_x
                psz = length_z / number_z
                return psz / psx
        else:
            return 1.0

    def getTotalDuration(self):
        """
        Method: Get total duration of the experiment using the .getDimensions() method
        Semi-private attribute `_duration`

        Return: duration (float) in seconds
        """
        if not hasattr(self, '_duration'):
            self._duration = 0.0
            for d in self.getDimensions():
                if dimName[int(d.getAttribute("DimID"))] == "T":
                    self._duration = float(d.getAttribute("Length"))
        return self._duration

    def getTimeLapse(self):
        """
        Method: Get an estimate of the average time lapse between two frames in seconds

        Return: estimated Lag time in seconds (float)

        warning:: Note that this value is accessible directly via getScannerSetting('nDelayTime_ms')

        """
        if self.getNbFrames() == 1:
            return 0
        else:
            return self.getTotalDuration() / (self.getNbFrames() - 1)

    def getTimeStamps(self):
        """
        Method: Get a numpy array of all image timeStamps in the Serie.
        Semi-private attribute `_timeStamps`

        Return: Numpy array of integers with timeStamps of all successives images in the Serie

        warning:: on first call getTimeStamps() suppresses the data from the XML SerieHeader
        """
        if not hasattr(self, '_timeStamps'):
            tslist = self.root.getElementsByTagName("TimeStampList")[0]
            if tslist.hasAttribute("NumberOfTimeStamps") and int(tslist.getAttribute("NumberOfTimeStamps")) > 0:
                # SP8 way of storing time stamps in the text of the node as 16bits hexadecimal separated by spaces
                self._timeStamps = np.array([
                    int(h, 16)
                    for h in tslist.firstChild.nodeValue.split()
                ])
            else:
                # SP5 way of storing time stamps as very verbose XML
                self._timeStamps = np.asarray([
                    (int(c.getAttribute("HighInteger")) << 32) + int(c.getAttribute("LowInteger"))
                    for c in tslist.getElementsByTagName("TimeStamp")])
                # remove the data from XML
                for c in tslist.getElementsByTagName("TimeStamp"):
                    c.parentNode.removeChild(c).unlink()
        return self._timeStamps

    def getRelativeTimeStamps(self):
        """
        Method: Get a numpy array of all image relativetimeStamps in the Serie.
        Semi-private attribute `_relTimeStamps`

        Return: Numpy array of integers with relativetimeStamps of all successives images in the Serie

        warning:: on first call getRelativeTimeStamps() suppresses the data from the XML SerieHeader
        """
        if not hasattr(self, '_relTimeStamps'):
            self._relTimeStamps = np.asarray([
                float(c.getAttribute("Time"))
                for c in self.root.getElementsByTagName("RelTimeStamp")])
            # remove the data from XML
            for c in self.root.getElementsByTagName("RelTimeStamp"):
                c.parentNode.removeChild(c).unlink()
        return self._relTimeStamps

    def getBytesInc(self, dimension):
        """
        Method: Get the ByteIncrement of a given dimension in the Serie.
        Semi-private attribute `_"dim"`, with "dim" a string = X, Y, Z, T

        arg:: dimension either as a string or as its key (integer) in dimName = {1: "X", 2: "Y", 3: "Z", 4: "T", ...}

        Return: BytesIncr as integer
        """
        # todo: consider channels
        if isinstance(dimension, int):
            dim = dimName[dimension]
        else:
            dim = dimension
        if not hasattr(self, '_' + dim):
            setattr(self, '_' + dim, 0)
            for d in self.getDimensions():
                if dimName[int(d.getAttribute("DimID"))] == dim:
                    setattr(self, '_' + dim, int(d.getAttribute("BytesInc")))
        return getattr(self, '_' + dim)

    def chooseChannel(self):
        """
        Method: Interactive selection of the channel for the current serie.

        Print the Serie name `chosen_serie.getName() and a list of all channels specifying: index, Color
        Promt for selecting a channel index.

        """
        st = "Serie: %s\n" % self.getName()
        for i, c in enumerate(self.getChannels()):
            st += "(%i) %s\n" % (i, channelTag[int(c.getAttribute("ChannelTag"))])
        print(st)
        if len(self.getChannels()) < 2:
            r = 0
        while (True):
            try:
                r = int(input("Choose a channel --> "))
                if r < 0 or r > len(self.getChannels()):
                    raise ValueError()
                break
            except ValueError:
                print("Oops!  That was no valid number.  Try again...")
        return r

    def getNbFrames(self):
        """
        Method: Get the number of frames in the Serie (acquisition at successive times)
        Semi-private attribute `_nbFrames`

        Return: number of frames (integer)
        """
        if not hasattr(self, '_nbFrames'):
            self._nbFrames = 1
            for d in self.getDimensions():
                if d.getAttribute("DimID") == "4":
                    self._nbFrames = int(d.getAttribute("NumberOfElements"))
        return self._nbFrames

    def getBoxShape(self):
        """
        Method: Get the Shape (spatial) of a frame
        Semi-private attribute: _boxShape

        Return: a list integers [length axis 1, length axis 2, ... ] ordered according to axis number (X, Y, Z)

        """
        if not hasattr(self, '_boxShape'):
            dims = {
                int(d.getAttribute('DimID')): int(d.getAttribute("NumberOfElements"))
                for d in self.getDimensions()
            }
            # ensure dimensions are sorted (unlike dictionnaries...), keep only spatial dimensions
            self._boxShape = [s for d, s in sorted(dims.items()) if d < 4]
        return self._boxShape

    def getFrameShape(self):
        """
        Method: Get the Shape of the frame (nD image) in C order, that is Z,Y,X

        Return: list of integers of axis length in Z, Y, X order (reverse as in .getBoxShape())
        """
        return self.getBoxShape()[::-1]

    def get2DShape(self):
        """
        Method: Get the Shape of an image using the two first spatial dimensions, in C order, e.g. Y,X

        Return: list of integers of axis length
        """
        return self.getBoxShape()[:2][::-1]

    def getNbPixelsPerFrame(self):
        """
        Method: Get the total number of pixels in a frame of shape .getBoxShape()
        Semi-private attribute: _nbPixelsPerFrame

        Return: total number of pixels (integer)
        """
        if not hasattr(self, '_nbPixelsPerFrame'):
            self._nbPixelsPerFrame = np.prod(self.getBoxShape())
        return self._nbPixelsPerFrame

    def getNbPixelsPerSlice(self):
        """
        Method: Get the total number of pixels in a Slice of shape .get2DShape()
        Semi-private attribute: _nbPixelsPerSlice

        Return: total number of pixels (integer)
        """
        if not hasattr(self, '_nbPixelsPerSlice'):
            self._nbPixelsPerSlice = np.prod(self.get2DShape())
        return self._nbPixelsPerSlice


def get_xml(lif_name: str):
    """
    Function: Extract the XML header from LIF file and save it. Generated .xml file can be opened in a Web Browser

    Use to examine the global architecture and the keywords associated with usefull information, to use with getXXX() methods
    """
    with open(lif_name, "rb") as f:
        memBlock, trash, testBlock = struct.unpack("iic", f.read(9))
        if memBlock != 112:
            raise Exception("This is not a valid LIF file")
        if testBlock != b'*':
            raise Exception("Invalid block at %l" % f.tell())
        memorysize, = struct.unpack("I", f.read(4))
        # read and parse the header
        xml = f.read(2 * memorysize).decode("utf-16")
        return xml


class Reader(Header):
    """
    Reads Leica LIF files

    Methods: getSeries(), chooseSeries(), __init__(), __iter__(), __readMemoryBlockHeader()

    Semi-Private Attribute:: _series


    """

    def __init__(self, lifFile, quick=True):
        
        # open file and find it's size
        if isinstance(lifFile, io.IOBase):
            self.f = lifFile
        else:
            self.f = open(lifFile, "rb")
        self.f.seek(0, 2)
        filesize = self.f.tell()
        self.f.seek(0)

        # read the size of the memory block containing the XML header
        # takes position at the begining of the XML header
        xmlHeaderLength = self.__readMemoryBlockHeader()
        #
        # xmlHeaderLength, = struct.unpack("L",self.f.read(4))

        # Read the XML header as raw buffer. It should avoid encoding problems
        # but who uses japanese characters anyway
        xmlHeaderString = self.f.read(xmlHeaderLength * 2).decode('latin-1')
        self.parse(io.StringIO(xmlHeaderString[::2]), quick)


        # Index the series offsets
        self.offsets = []
        while (self.f.tell() < filesize):

            memorysize = self.__readMemoryBlockHeader()

            while (self.f.read(1) != b"*"):
                pass
            # size of the memory description
            memDescrSize, = struct.unpack("I", self.f.read(4))
            memDescrSize *= 2
            # skip the description: we are at the begining of the content
            self.f.seek(memDescrSize, 1)
            # add image offset if memory size >0
            if memorysize > 0:
                self.offsets.append(self.f.tell())
                self.f.seek(memorysize, 1)
        if not quick:
            # convert immediately the time stamps in XML format to lighweight numpy array
            for s in self:
                s.getTimeStamps()
                s.getRelativeTimeStamps()

        # self.offsets = [int(m.getAttribute("Size")) for m in self.xmlHeader.getElementsByTagName("Memory")]
        # self.offsets = [i for i in self.offsets if i != 0]


    def __readMemoryBlockHeader(self):
        memBlock, trash, testBlock = struct.unpack("iic", self.f.read(9))
        if memBlock != 112:
            raise Exception("This is not a valid LIF file")
        if testBlock != b'*':
            raise Exception("Invalid block at %d" % self.f.tell())
        if not hasattr(self, 'xmlHeader') or self.getVersion() < 2:
            memorysize, = struct.unpack("I", self.f.read(4))
        else:
            memorysize, = struct.unpack("Q", self.f.read(8))
        return memorysize

    def getSeries(self):
        """
        Method: Get the experimental Series from the raw .lif file
        Semi-Private Attribute: _series

        Return: a List of class Serie objects
        """
        if not hasattr(self, '_series'):
            self._series = [
                Serie(s.root, self.f, self.offsets[i]) for i, s in enumerate(self.getSeriesHeaders())
            ]
        return self._series

    def chooseSerie(self):
        """
        Method: use .chooseSerieIndex() inherited Header method to interactively choose a Serie using .getSeries()

        Return: the selected class Serie object
        """
        return self.getSeries()[self.chooseSerieIndex()]

    def __iter__(self):
        return iter(self.getSeries())


class Serie(SerieHeader):
    """
    One of the datasets (Serie) in a .lif file

    Methods:
    """

    def __init__(self, serieElement, f, offset):
        self.f = f
        self.__offset = offset
        self.root = serieElement

    def getOffset(self, **dimensionsIncrements):
        """
        Method: Get the Frame Offset

        kwargs:: Time, Channel or data type. Default to `channel=0, T=0, dtype=np.uint8`
        """
        of = 0
        for d, b in dimensionsIncrements.items():
            of += self.getBytesInc(d) * b

        if of >= self.getMemorySize():
            raise IndexError("offset out of bound")
        return self.__offset + of

    def getChannelOffset(self, channel):
        """
        Method: Get the Channel Offset

        arg:: channel (int)
        """
        channels = self.getChannels()
        channel_node = channels[channel]
        of = int(channel_node.getAttribute('BytesInc'))
        return of

    def get2DSlice(self, **dimensionsIncrements):
        """
        Method: Use the two first dimensions as image dimension (XY, XZ, YZ). Axis are in C order (last index is X).

        Return: Image as numpy array with the axis in ZY, ZX, or YX order

        warning:: See dtype argument; might not support 16bits encoding
        """
        for d in self.getDimensions()[:2]:
            if dimName[int(d.getAttribute("DimID"))] in dimensionsIncrements:
                raise Exception('You can\'t set %s in serie %s' % (
                    dimName[int(d.getAttribute("DimID"))],
                    self.getName())
                                )

        self.f.seek(self.getOffset(**dimensionsIncrements))
        shape = self.get2DShape()
        return np.fromfile(
            self.f,
            dtype=np.ubyte,
            count=self.getNbPixelsPerSlice()
        ).reshape(shape)

    def get2DString(self, **dimensionsIncrements):
        """
        Use the two first dimensions as image dimension
        """
        for d in self.getDimensions()[:2]:
            if dimName[int(d.getAttribute("DimID"))] in dimensionsIncrements:
                raise Exception('You can\'t set %s in serie %s' % (
                    dimName[int(d.getAttribute("DimID"))],
                    self.getName()))

        self.f.seek(self.getOffset(**dimensionsIncrements))
        return self.f.read(self.getNbPixelsPerSlice())

    def getFrame(self, T=0, channel=0, dtype='uint16'):
        """
        Return a numpy array (C order, thus last index is X):
         2D if XYT or XZT serie,
         3D if XYZ, XYZT or XZYT
         (ok if no T dependence)
        Leica use uint8 by default, but after deconvolution the datatype is np.uint16 NOT ALWAYS TRUE - but maybe true enough???

        """
        zyx = np.zeros(self.getFrameShape(), dtype='uint16')

        channels = self.getChannels()

        for z in range(self.getFrameShape()[0]):
            channel_offset = self.getChannelOffset(channel)

            # For whatever reason, the blue channel offset is off by one. WTF
            if channel == 0:
                channel_offset += 1
            self.f.seek(self.getOffset(T=T, Z=z) + channel_offset)
            zyx[z, ...] = np.fromfile(self.f, dtype=dtype, count=int(self.getNbPixelsPerSlice())).reshape(self.get2DShape())

        if zyx.min() < 0:
            raise ValueError(zyx.min() >= 0, zyx.min(), zyx.max(), zyx.dtype, type(zyx))

        return zyx

    def getFrame2D(self, channel=0, T=0, dtype=np.uint8):
        """
        Method: Get a 2D image from the serie XY (for XY, XYT) or XZ (for XZ, XZT)

        kwarg:: channel number (int, default = 0), Time index (int, default = 0)

        Return: a 2D numpy array (in C order: last index is X). (ok if no T dependence)
                Leica use uint8 by default, but after deconvolution the datatype is np.uint16
        """
        channels = self.getChannels()
        cyx = []
        for i in range(len(channels)):
            self.f.seek(self.getOffset(**dict({'T': T})) + self.getChannelOffset(i))
            yx = np.fromfile(self.f, dtype=dtype, count=int(self.getNbPixelsPerSlice()))
            yx = yx.reshape(self.get2DShape())
            cyx.append(yx)
        cyx = np.array(cyx)
        return cyx[channel]

    def getXYC(self, T=0, dtype=np.uint8):
        """
        Get a 3D image whose shape is (X, Y, channel_number).
            This is typically used to read a multiple-channeled 2D scan

        Args:
            T (int): The frame number (time index).
            dtype (type): the datatype of the pixel intensities.
                Leica use uint8 by default, but after the deconvolution
                the datatype is np.uint16.

        Return:
            np.ndarray: 3D array shape:  (X, Y, channel_number).
        """
        n_channels = len(self.getChannels())
        shape = [n_channels] + list(self.get2DShape())
        print(shape)


        print(self.getOffset(T=T))

        self.f.seek(self.getOffset(**dict({'T': T})))

        xyc = np.fromfile(
            self.f, dtype=dtype,
            count=int(self.getNbPixelsPerSlice()) * n_channels
        )

        # shape = [shape[i] for i in [1,0,2]]
        xyc = xyc.reshape(shape)
        return xyc

    def getXYZ(self, T=0, channel=0, dtype=np.uint8):
        """
        Get a 3D image whose shape is (X, Y, Z).
            This is typically used to read a 3D confocal image.

        Args:
            T (int): The frame number (time index), default value is 0.
            channel (int): the channel index, default value is 0.
            dtype (type): the datatype of the pixel intensities.
                Leica use uint8 by default, but after the deconvolution
                the datatype is np.uint16.

        Return:
            np.ndarray: 3D array shape:  (X, Y, Z).
        """
        zyx = self.getFrame(T=T, channel=channel, dtype=dtype)
        yxz = np.moveaxis(zyx, 0, -1)
        xyz = np.moveaxis(yxz, 0, 1)
        return xyz

    def getMetadata(self):
        """
        voxel size unit: µm
        """
        nbx, nby, nbz = self.getBoxShape()
        setting_records = self.root.getElementsByTagName('ScannerSettingRecord')
        dimension_descriptions = self.root.getElementsByTagName('DimensionDescription')
        if setting_records:
            # ScannerSettingRecord only available for some lif files!
            psx = self.getVoxelSize(1)  # m ---> µm
            psy = self.getVoxelSize(2)  # m ---> µm
            psz = self.getVoxelSize(3)  # m ---> µm
            unit_x = [s.getAttribute('Unit') for s in setting_records if s.getAttribute('Identifier') == 'dblVoxelX'][0]
            unit_y = [s.getAttribute('Unit') for s in setting_records if s.getAttribute('Identifier') == 'dblVoxelY'][0]
            unit_z = [s.getAttribute('Unit') for s in setting_records if s.getAttribute('Identifier') == 'dblVoxelZ'][0]
            units = [unit_x, unit_y, unit_z]
        elif dimension_descriptions:
            # Use DimensionDescription to get voxel information
            length_x = float(
                [d.getAttribute('Length') for d in dimension_descriptions if d.getAttribute('DimID') == '1'][0])
            length_y = float(
                [d.getAttribute('Length') for d in dimension_descriptions if d.getAttribute('DimID') == '2'][0])
            length_z = float(
                [d.getAttribute('Length') for d in dimension_descriptions if d.getAttribute('DimID') == '3'][0])
            number_x = float(
                [d.getAttribute('NumberOfElements') for d in dimension_descriptions if d.getAttribute('DimID') == '1'][
                    0])
            number_y = float(
                [d.getAttribute('NumberOfElements') for d in dimension_descriptions if d.getAttribute('DimID') == '2'][
                    0])
            number_z = float(
                [d.getAttribute('NumberOfElements') for d in dimension_descriptions if d.getAttribute('DimID') == '3'][
                    0])
            psx = length_x / number_x
            psy = length_y / number_y
            psz = length_z / number_z
            units = [s.getAttribute('Unit') for s in dimension_descriptions]
        else:
            raise RuntimeError("Can't find voxel information in the lif file!")

        if len(set(units)) == 1 and 'm' in units:
            factor = 1e6
            unit = 'um'
        else:
            warnings.warn('unit is not meter, check the unit of voxel size')
            factor = 1
            unit = ", ".join(units)

        psx = psx * factor  # m ---> µm
        psy = psy * factor  # m ---> µm
        psz = psz * factor  # m ---> µm

        return {
            'voxel_size_x': psx,
            'voxel_size_y': psy,
            'voxel_size_z': psz,
            'voxel_size_unit': unit,
            'voxel_number_x': nbx,
            'voxel_number_y': nby,
            'voxel_number_z': nbz,
            'channel_number': len(self.getChannels()),
            'frame_number': self.getNbFrames(),
        }


    def getMetadata2D(self):
        """
        voxel size unit: µm, working with a 2D slice
        """
        nbx, nby = self.getBoxShape()
        setting_records = self.root.getElementsByTagName('ScannerSettingRecord')
        dimension_descriptions = self.root.getElementsByTagName('DimensionDescription')
        if setting_records:
            # ScannerSettingRecord only available for some lif files!
            psx = self.getVoxelSize(1)  # m ---> µm
            psy = self.getVoxelSize(2)  # m ---> µm
            unit_x = [s.getAttribute('Unit') for s in setting_records if s.getAttribute('Identifier') == 'dblVoxelX'][0]
            unit_y = [s.getAttribute('Unit') for s in setting_records if s.getAttribute('Identifier') == 'dblVoxelY'][0]
            units = [unit_x, unit_y]
        elif dimension_descriptions:
            # Use DimensionDescription to get voxel information
            length_x = float(
                [d.getAttribute('Length') for d in dimension_descriptions if d.getAttribute('DimID') == '1'][0])
            length_y = float(
                [d.getAttribute('Length') for d in dimension_descriptions if d.getAttribute('DimID') == '2'][0])
            number_x = float(
                [d.getAttribute('NumberOfElements') for d in dimension_descriptions if d.getAttribute('DimID') == '1'][
                    0])
            number_y = float(
                [d.getAttribute('NumberOfElements') for d in dimension_descriptions if d.getAttribute('DimID') == '2'][
                    0])
            psx = length_x / number_x
            psy = length_y / number_y
            units = [s.getAttribute('Unit') for s in dimension_descriptions]
        else:
            raise RuntimeError("Can't find voxel information in the lif file!")

        if len(set(units)) == 1 and 'm' in units:
            factor = 1e6
            unit = 'um'
        else:
            warnings.warn('unit is not meter, check the unit of voxel size')
            factor = 1
            unit = ", ".join(units)

        psx = psx * factor  # m ---> µm
        psy = psy * factor  # m ---> µm

        return {
            'voxel_size_x': psx,
            'voxel_size_y': psy,
            'voxel_size_unit': unit,
            'voxel_number_x': nbx,
            'voxel_number_y': nby,
            'channel_number': len(self.getChannels()),
            'frame_number': self.getNbFrames(),
        }

    def saveXML(self, name=None):
        if not name:
            name = '{}.xml'.format(self.getName())
        with open(name, 'w') as f:
            f.write(self.root.toprettyxml())

    def getVTK(self, fname, T=0):
        """
        Export the frame at time T to a vtk file
        One can render the vtk file using software like Paraview
        """
        if '.vtk' not in fname:
            fname += '.vtk'
        with open(fname, 'wb') as f:
            header = '# vtk DataFile Version 3.0\n%s\n' % self.getName +\
                'BINARY\nDATASET STRUCTURED_POINTS\n' +\
                ('DIMENSIONS %d %d %d\n' % tuple(self.getBoxShape())) +\
                'ORIGIN 0 0 0\n' +\
                ('SPACING 1 1 %g\n' % self.getZXratio()) +\
                ('POINT_DATA %d\n' % self.getNbPixelsPerFrame()) +\
                'SCALARS Intensity unsigned_char\nLOOKUP_TABLE default\n'
            f.write(str.encode(header))
            self.f.seek(self.getOffset(**dict({'T': T})))
            f.write(self.f.read(self.getNbPixelsPerFrame()))

    def enumByFrame(self):
        """yield time steps one after the other as a couple (time,numpy array).
        It is not safe to combine this syntax with getFrame or get2DSlice.

        This implementation works only for single channel, 8 bytes series.
        """
        self.f.seek(self.getOffset())
        for t in range(self.getNbFrames()):
            yield t, np.fromfile(
                self.f,
                dtype=np.ubyte,
                count=self.getNbPixelsPerFrame()
            ).reshape(self.getFrameShape())

    def enumBySlice(self):
        """yield 2D slices one after the other as a 3-tuple (time,z,numpy array).
        It is not safe to combine this syntax with getFrame or get2DSlice.

        This implementation works only for single channel, 8 bytes series.
        """
        self.f.seek(self.getOffset())
        for t in range(self.getNbFrames()):
            for z in range(self.getNbPixelsPerFrame() / self.getNbPixelsPerSlice()):
                yield t, z, np.fromfile(
                    self.f,
                    dtype=np.ubyte,
                    count=self.getNbPixelsPerSlice()
                ).reshape(self.get2DShape())



import torch
from torch import Tensor
from typing import List, Tuple, Optional, Union, Dict

import torchvision.transforms.functional
from torchvision.transforms.functional import gaussian_blur

from hcat.lib.explore_lif import Reader
from hcat.train.transforms import _crop

import numpy as np
import skimage.io as io
import os.path

from tqdm import trange

import matplotlib.pyplot as plt
import xml.etree.ElementTree as ET


def graceful_exit(message):
    """ Decorator which returns a message upon failure"""

    def decorator(function):
        def wrapper(*args, **kwargs):
            result = None
            try:
                result = function(*args, **kwargs)
            except Exception:
                print(message)
            return result

        return wrapper

    return decorator


def calculate_indexes(pad_size: int, eval_image_size: int,
                      image_shape: int, padded_image_shape: int) -> List[List[int]]:
    """
    This calculates indexes for the complete evaluation of an arbitrarily large image by unet.
    each index is offset by eval_image_size, but has a width of eval_image_size + pad_size * 2.
    Unet needs padding on each side of the evaluation to ensure only full convolutions are used
    in generation of the final mask. If the algorithm cannot evenly create indexes for
    padded_image_shape, an additional index is added at the end of equal size.

    :param pad_size: int corresponding to the amount of padding on each side of the
                     padded image
    :param eval_image_size: int corresponding to the shape of the image to be used for
                            the final mask
    :param image_shape: int Shape of image before padding is applied
    :param padded_image_shape: int Shape of image after padding is applied

    :return: List of lists corresponding to the indexes
    """

    # We want to account for when the eval image size is super big, just return index for the whole image.
    if eval_image_size + (2 * pad_size) > image_shape:
        return [[0, image_shape - 1]]

    try:
        ind_list = torch.arange(0, image_shape, eval_image_size)
    except RuntimeError:
        raise RuntimeError(f'Calculate_indexes has incorrect values {pad_size} | {image_shape} | {eval_image_size}:\n'
                           f'You are likely trying to have a chunk smaller than the set evaluation image size. '
                           'Please decrease number of chunks.')
    ind = []
    for i, z in enumerate(ind_list):
        if i == 0:
            continue
        z1 = int(ind_list[i - 1])
        z2 = int(z - 1) + (2 * pad_size)
        if z2 < padded_image_shape:
            ind.append([z1, z2])
        else:
            break
    if not ind:  # Sometimes z is so small the first part doesnt work. Check if z_ind is empty, if it is do this!!!
        z1 = 0
        z2 = eval_image_size + pad_size * 2
        ind.append([z1, z2])
        ind.append([padded_image_shape - (eval_image_size + pad_size * 2), padded_image_shape])
    else:  # we always add at the end to ensure that the whole thing is covered.
        z1 = padded_image_shape - (eval_image_size + pad_size * 2)
        z2 = padded_image_shape - 1
        ind.append([z1, z2])
    return ind


@torch.jit.script
def crop_to_identical_size(a: torch.Tensor, b: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Crops Tensor a to the shape of Tensor b, then crops Tensor b to the shape of Tensor a.

    :param a: torch.
    :param b:
    :return:
    """
    # if a.ndim != b.ndim:
    #     raise RuntimeError('Number of dimensions of tensor "a" does not equal tensor "b".')

    if a.ndim < 3:
        raise RuntimeError('Only supports tensors with minimum 3dimmensions and shape [..., X, Y, Z]')

    a = _crop(a, x=0, y=0, z=0, w=b.shape[-3], h=b.shape[-2], d=b.shape[-1])
    b = _crop(b, x=0, y=0, z=0, w=a.shape[-3], h=a.shape[-2], d=a.shape[-1])
    return a, b


# ########################################################################################################################
# #                                                       Postprocessing
# ########################################################################################################################

# @graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not remove edge cells.' + '\x1b[0m')
@torch.jit.script
def remove_edge_cells(mask: torch.Tensor) -> torch.Tensor:
    """
    Removes cells touching the border

    .. warning:
       Will not raise an error upon failure, instead returns None and prints to standard out

    :param mask: (B, X, Y, Z)
    :return: mask (B, X, Y, Z)
    """

    if mask.ndim != 4: raise RuntimeError('input.ndim != 4')

    # Mask is empty, nothing to do.
    if mask.max() == 0:
        return mask

    left = torch.unique(mask[:, 0, :, :])
    right = torch.unique(mask[:, -1, :, :])
    top = torch.unique(mask[:, :, 0, :])
    bottom = torch.unique(mask[:, :, -1, :])

    cells = torch.unique(torch.cat((left, right, top, bottom)))

    for c in cells:
        if c == 0:
            continue
        mask[mask == c] = 0

    return mask


# @graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not remove improperly sized cells.' + '\x1b[0m')
@torch.jit.script
def remove_wrong_sized_cells(mask: torch.Tensor) -> torch.Tensor:
    """
    Removes cells with outlandish volumes. These have to be wrong.

    .. warning:
       Will not raise an error upon failure, instead returns None and prints to standard out

    :param mask: [B, C=1, X, Y, Z] int torch.Tensor: cell segmentation mask where each cell has a unique cell id
    :return:
    """
    unique = torch.unique(mask)
    unique = unique[unique.nonzero()]

    for u in unique:
        if (mask == u).sum() < 4000:
            mask[mask == u] = 0
        elif (mask == u).sum() > 30000:
            mask[mask == u] = 0
    return mask


# ########################################################################################################################
# #                                                       U Net Specific
# ########################################################################################################################


def pad_image_with_reflections(image: torch.Tensor, pad_size: Tuple[int] = (30, 30, 6)) -> torch.Tensor:
    """
    Pads image according to Unet spec
    expect [B, C, X, Y, Z]
    Adds pad size to each side of each dim. For example, if pad size is 10, then 10 px will be added on top, and on bottom.

    :param image:
    :param pad_size:
    :return:
    """
    if not isinstance(image, torch.Tensor):
        raise TypeError(f'Expected image to be of type torch.tensor not {type(image)}')
    for pad in pad_size:
        if pad % 2 != 0:
            raise ValueError('Padding must be divisible by 2')

    device = image.device

    image_size = image.shape
    pad_size = np.array(pad_size)

    left_pad = image.cpu().numpy()[:, :, pad_size[0] - 1::-1, :, :]
    left_pad = torch.as_tensor(left_pad.copy())
    right_pad = image.cpu().numpy()[:, :, -1:-pad_size[0] - 1:-1, :, :]
    right_pad = torch.as_tensor(right_pad.copy())
    image = torch.cat((left_pad, image.cpu(), right_pad), dim=2).to(device)

    left_pad = 0
    right_pad = 0

    bottom_pad = image.cpu().numpy()[:, :, :, pad_size[1] - 1::-1, :]
    bottom_pad = torch.as_tensor(bottom_pad.copy())
    top_pad = image.cpu().numpy()[:, :, :, -1:-pad_size[1] - 1:-1, :]
    top_pad = torch.as_tensor(top_pad.copy())
    image = torch.cat((bottom_pad, image.cpu(), top_pad), dim=3).to(device)
    bottom_pad = 0
    top_pad = 0

    bottom_pad = image.cpu().numpy()[:, :, :, :, pad_size[2] - 1::-1]
    bottom_pad = torch.as_tensor(bottom_pad.copy())
    top_pad = image.cpu().numpy()[:, :, :, :, -1:-pad_size[2] - 1:-1]
    top_pad = torch.as_tensor(top_pad.copy())

    return torch.cat((bottom_pad, image.cpu(), top_pad), dim=4).to(device)


########################################################################################################################
#                                                         Generics
########################################################################################################################

def prep_dict(data_dict, device: str):
    if isinstance(data_dict, dict):
        images = data_dict['image'].to(device).float()
        data_dict['boxes'] = data_dict['boxes']
        data_dict['labels'] = data_dict['labels']
        data_dict = [data_dict]
    elif isinstance(data_dict, list):
        images = []
        for dd in data_dict:
            images.append(dd['image'].to(device).float().squeeze(0))
        for i in range(len(data_dict)):
            data_dict[i]['boxes'] = data_dict[i]['boxes']
            data_dict[i]['labels'] = data_dict[i]['labels']
    else:
        raise RuntimeError('data_dict is neither list nor dict')

    return images, data_dict


def warn(message: str, color: str) -> None:
    c = {'green': '\x1b[1;32;40m',
         'yellow': '\x1b[1;33;40m',
         'red': '\x1b[1;31;40m',
         'norm': '\x1b[0m'}  # green, yellow, red, normal

    print(c[color.lower()] + message + c['norm'])


def get_device(verbose: Optional[bool] = False) -> str:
    if verbose: print(f'[      ] Initializing Hardware Accelerator...', end='')
    if torch.cuda.is_available():
        device = 'cuda:0'
        if verbose:
            print("\r[\x1b[1;32;40m CUDA \x1b[0m]")
    elif torch.backends.mps.is_available():
        device = 'mps'
        if verbose:
            print("\r[\x1b[1;32;40m MPS  \x1b[0m]")
    else:
        device = 'cpu'
        if verbose:
            print("\r[\x1b[1;33;40m CPU  \x1b[0m]")
            warn('WARNING: GPU not present or CUDA is not correctly initialized for GPU accelerated computation. '
                 'Analysis may be slow.', color='yellow')

    return device


def load(file: str, header_name: Optional[str] = 'TileScan 1 Merged',
         verbose: bool = False, dtype: str ='uint16', ndim: int = 3) -> Union[None, np.array]:
    """
    Loads image file (*leica or *tif) and returns an np.array

    :param file: str path to the file
    :param header_name: Optional[str] header name of lif. Does nothing image file is a tif
    :param verbose: bool print status of image loading to standard out.
    :return: np.array image matrix from file, aborts if the image is too large and returns None.
    """

    image_base = None

    if verbose:
        print(f'[      ] Loading {file}...', end='')

    if not os.path.exists(file):
        print(f'\n\x1b[1;31;40m' + f'Cannot access: \'{file}\'. No such file.' + '\x1b[0m')
        return None

    if file.endswith('.lif'):  # Load lif file
        reader = Reader(file)
        series = reader.getSeries()
        for i, header in enumerate(reader.getSeriesHeaders()):
            if header_name in header.getName():  ###'TileScan 1 Merged':

                chosen = series[i]
                for c in trange(4, desc=f'Loading LIF Image with header: {header_name}'):
                    if c == 0 and ndim == 3:
                        image_base = chosen.getXYZ(T=0, channel=c, dtype=dtype)[np.newaxis]

                        if image_base.size * 4 > 9000 * 9000 * 40 * 4:
                            print(
                                f'\x1b[1;33;40m' + f'WARNING: \'{file}\' has image size of {image_base.shape} and is very large. Analysis may fail.' + '\x1b[0m')
                            return None

                    else:
                        image_base = np.concatenate((image_base, chosen.getXYZ(T=0, channel=c)[np.newaxis]), axis=0)

            # del series, header, chosen

    elif file.endswith('.tif') or file.endswith('.png') or file.endswith('.jpg'):  # Load a tif
        image_base = io.imread(file)

        if image_base.ndim == 4:
            image_base = image_base.transpose((-1, 1, 2, 0))

        elif image_base.ndim == 3 and np.array(
                image_base.shape).min() > 4:  # Suppose you load a 3D image with one channel.
            image_base = image_base[np.newaxis, ...]
            image_base = np.concatenate((image_base, image_base, image_base, image_base), axis=0).transpose(0, 2, 3, 1)

        elif image_base.ndim == 3 and image_base.shape[-1] <= 4:  # Suppose you load a 2D image! with multiple channels

            image_base = image_base[...].transpose((2, 0, 1))  # always just take the last two channels
            # just return it...
            pass
        elif image_base.ndim == 3 and image_base.shape[0] <= 4:  # Suppose you load a 2D image! with multiple channels
            pass
        else:
            print(
                f'\x1b[1;31;40m' + f'Cannot load: \'{file}\'. Unsupported number of dimmensions: {image_base.ndim}' + '\x1b[0m')
            return None

    else:
        print(f'\x1b[1;31;40m' + f'Cannot load: \'{file}\'. Filetype not supported.' + '\x1b[0m')
        return None

    if verbose:
        print("\r[\x1b[1;32;40m DONE \x1b[0m]")
    return image_base


def rescale_box_sizes(boxes: torch.Tensor,
                      current_pixel_size: Optional[float] = None,
                      cell_diameter: Optional[int] = None,
                      ):
    scale = 1.0

    if current_pixel_size is not None:
        scale = current_pixel_size / 288.88

    elif cell_diameter is not None:
        scale = 30 / cell_diameter

    return boxes / scale


def correct_pixel_size_image(image: torch.Tensor,
                             current_pixel_size: Optional[float] = None,
                             cell_diameter: Optional[int] = None,
                             antialias: Optional[bool] = True,
                             verbose: Optional[bool] = False):
    """
    Correct an image to new pixel size...

    :param image:
    :param cell_diameter:
    :param current_pixel_size:
    :param antialias:
    :param verbose:
    :return:
    """
    if cell_diameter is None and current_pixel_size is None:
        print(f'Image was not scaled. Assuming pixel size of 288.88nm X/Y')
        return image
    else:
        if verbose:
            print(f'[      ] Scaling Image...', end='')

    if current_pixel_size is not None:
        optimal_pixel_size = 288.88
        scale = current_pixel_size / optimal_pixel_size
        c, x, y = image.shape
        new_size = [round(x * scale), round(y * scale)]

    elif cell_diameter is not None:
        scale = 30 / cell_diameter
        c, x, y = image.shape
        new_size = [round(x * scale), round(y * scale)]

    if image.min() < 0:
        image = image.div(0.5).add(0.5)
        image = torchvision.transforms.functional.resize(image, new_size, antialias=antialias).mul(0.5).add(0.5)
    else:
        image = torchvision.transforms.functional.resize(image, new_size, antialias=antialias)

    if verbose:
        print("\r[\x1b[1;32;40m DONE \x1b[0m]")

    return image


def get_dtype_offset(dtype: str = 'uint16', image_max=None) -> int:
    """ get dtype from string """

    encoding = {
        'uint16': 2 ** 16,
        'uint8': 2 ** 8,
        'uint12': 2 ** 12,
        'float64': 1,
    }
    dtype = str(dtype)

    if image_max is not None and dtype=='uint16':
        dtype = 'uint12' if 2**12 + 1 > image_max else 'uint16'

    if dtype in encoding:
        scale = encoding[dtype]

    else:
        print(
            f'\x1b[1;31;40m' + f'ERROR: Unsupported dtype: {dtype}. Currently support: {[k for k in encoding]}' + '\x1b[0m')
        scale = None
        if image_max:
            print(f'Appropriate Scale Factor inferred from image maximum: {image_max}')
            if image_max <= 256:
                scale = 256
            else:
                scale = image_max
    return scale


def cochlea_to_xml(cochlea, filename=None) -> None:
    # Xml header and root
    if filename is None:
        filename = cochlea.filename

    path = cochlea.path  # has filename appended to end
    folder = os.path.split(filename)[0]
    folder = os.path.split(folder)[-1] if len(folder) != 0 else os.path.split(os.getcwd())[1]

    _, height, width = cochlea.im_shape
    depth = 1

    root = ET.Element('annotation')
    ET.SubElement(root, 'folder').text = folder
    ET.SubElement(root, 'filename').text = filename
    ET.SubElement(root, 'path').text = path
    source = ET.SubElement(root, 'source')
    ET.SubElement(source, 'database').text = 'unknown'
    size = ET.SubElement(root, 'size')
    ET.SubElement(size, 'width').text = str(width)
    ET.SubElement(size, 'height').text = str(height)
    ET.SubElement(size, 'depth').text = str(depth)
    ET.SubElement(root, 'segmented').text = '0'

    for c in cochlea.cells:
        x0, y0, x1, y1 = c.boxes
        #  xml write xmin, xmax, ymin, ymax
        object = ET.SubElement(root, 'object')
        ET.SubElement(object, 'name').text = c.type
        ET.SubElement(object, 'pose').text = 'Unspecified'
        ET.SubElement(object, 'truncated').text = '0'
        ET.SubElement(object, 'difficult').text = '0'
        bndbox = ET.SubElement(object, 'bndbox')
        ET.SubElement(bndbox, 'xmin').text = str(int(x0))
        ET.SubElement(bndbox, 'ymin').text = str(int(y0))
        ET.SubElement(bndbox, 'xmax').text = str(int(x1))
        ET.SubElement(bndbox, 'ymax').text = str(int(y1))

    tree = ET.ElementTree(root)
    filename = os.path.splitext(cochlea.path)[0]
    tree.write(filename + '.xml')


def normalize_image(image: Tensor, normalize: bool, verbose: Optional[bool] = False) -> Tensor:
    if verbose and normalize:
        print(f'[      ] Normalizing to maximum brightness...', end='')
    for c in range(image.shape[0]):
        max_pixel = gaussian_blur(image[c, ...].unsqueeze(0), kernel_size=[7, 7], sigma=0.5).max()
        image[c, ...] = image[c, ...].div(max_pixel + 1e-16).clamp(0, 1) if max_pixel != 0 else image[
            c, ...]
    if verbose and normalize:
        print("\r[\x1b[1;32;40m DONE \x1b[0m]")

    return image

def make_rgb(image: Tensor) -> Tensor:
    _, x, y = image.shape
    if image.shape[0] == 2:
        image: Tensor = torch.cat((torch.zeros((1, x, y), device=image.device), image), dim=0)

    return image[0:3, ...] #ALWAYS choose


def image_to_float(image: Union[np.ndarray, Tensor], scale: int, verbose: Optional[bool] = False) -> Tensor:
    if verbose:
        print(f'[      ] Converting Image to Float... ', end='')

    if isinstance(image, np.ndarray):
        image = torch.from_numpy(image.astype(np.uint16) / scale).to(torch.float32)
    else:
        image = image.div(scale).to(torch.float32)

    if verbose:
        print("\r[\x1b[1;32;40m DONE \x1b[0m]")

    return image

def save_image_as_png(image: Tensor, filename: str, verbose: Optional[bool] = False) -> None:
    if verbose:
        print(f'[      ] Saving as {filename[:-4:]}.png...', end='')

    png = image.cpu().permute(1, 2, 0).mul(255).int().numpy().clip(0, 255).astype(np.uint8)
    io.imsave(filename[:-4:] + '.png', png[:, :, [2, 0, 1]])
    io.imsave(filename[:-4:] + '_scaled.tif', png[:, :, [2, 0, 1]])

    del png
    if verbose:
        print("\r[\x1b[1;32;40m DONE \x1b[0m]")


########################################################################################################################
#                                                         Plotting
########################################################################################################################

def plot_embedding(embedding: torch.Tensor, centroids: torch.Tensor) -> None:
    num = 25
    x = embedding.detach().cpu().numpy()[0, 0, ...].flatten() * num
    y = embedding.detach().cpu().numpy()[0, 1, ...].flatten() * num
    plt.hist2d(y, x, bins=(embedding.shape[2], embedding.shape[3]))
    print(centroids)
    plt.plot(centroids[0, :, 1].cpu().numpy(), centroids[0, :, 0].cpu().numpy(), 'ro')

    plt.show()

def make_embedding_image(embedding):
    x = embedding.detach().cpu().numpy()[0, 0, ...].flatten()
    y = embedding.detach().cpu().numpy()[0, 1, ...].flatten()
    histogram,_,_ = np.histogram2d(x, y, bins=(embedding.shape[2], embedding.shape[3]))
    return torch.from_numpy(histogram)



if __name__ == "__main__":
    cochlea = torch.load('test.cochlea')
    cochlea_to_xml(cochlea)



import torch
import torch.nn as nn
from warnings import filterwarnings
from typing import Tuple, Union

filterwarnings("ignore", category=UserWarning)

class StackedDilation(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel: Union[int, Tuple[int, int, int]],
                 dilation: Union[int, Tuple[int, int, int]] = None,
                 padding: Union[int, Tuple[int, int, int]] = None
                 ):

        super(StackedDilation, self).__init__()

        self.dilation = dilation if dilation is not None else (1, 2, 3, 4)
        self.padding = padding if padding is not None else (2, 4, 6, 8)

        self.conv = nn.Conv3d(in_channels, out_channels, kernel_size=kernel, dilation=1, padding=2)
        self.out_conv = nn.Conv3d(out_channels*len(self.dilation), out_channels, kernel_size=1, padding=0)
        self.activation = nn.ReLU()
        self.batch_norm_conv = nn.BatchNorm3d(out_channels)
        self.batch_norm_out = nn.BatchNorm3d(out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for i, (dilation, padding) in enumerate(zip(self.dilation, self.padding)):
            self.conv.dilation = dilation
            self.conv.padding = padding
            x = self.activation(self.batch_norm_conv(self.conv(x)))
            if i == 0:
                out = x
            else:
                out = torch.cat((out, x), dim=1)
        out = self.activation(self.batch_norm_out(self.out_conv(out)))
        return out



import torch
from torchvision.ops.boxes import box_convert
from typing import Tuple, Dict, Optional
from hcat.train.transforms import _crop


# DOCUMENTED

@torch.jit.script
def crop_to_identical_size(a: torch.Tensor, b: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Crops Tensor a to the shape of Tensor b, then crops Tensor b to the shape of Tensor a.


    :param a: input 1
    :param b: input 2
    :raises RuntimeError: If n_dim of tensor a and b are different
    :return:
    """
    if a.ndim != b.ndim:
        raise RuntimeError('Number of dimensions of tensor "a" does not equal tensor "b".')

    a = _crop(a, x=0, y=0, z=0, w=b.shape[2], h=b.shape[3], d=b.shape[4])
    b = _crop(b, x=0, y=0, z=0, w=a.shape[2], h=a.shape[3], d=a.shape[4])
    return a, b


@torch.jit.script
def _stat(asses: torch.Tensor, index: int) -> Dict[str, torch.Tensor]:
    """
    Calculate various statistics from an input cell tensor.

    :param asses: [C, 1] flattened input image.
    :param index: channel index to run analysis on.
    :return:
    """
    x = asses[index, :]
    numel = x.numel()

    return {'mean': x.mean(), 'median': torch.median(x),
            'std': x.std(), 'var': x.var(), 'min': x.min(), 'max': x.max(),
            '%saturated': torch.sum(x == 1).div(numel), '%zero': torch.sum(x == 0).div(numel)}


class Cell:
    def __init__(self,
                 image: Optional[torch.Tensor],
                 mask: Optional[torch.Tensor],
                 loc: torch.Tensor,
                 id: Optional[int] = None,
                 scores: Optional[torch.Tensor] = None,
                 boxes: Optional[torch.Tensor] = None,
                 cell_type: Optional[str] = None,
                 channel_name: Optional[Tuple[str]] = ('dapi', 'gfp', 'myo7a', 'actin')):
        """
        Dataclass of a single detected cell object.

        :param image: [B, C, X, Y ,Z] image crop of a *single* cell
        :param mask: [B, C, X, Y ,Z] segmentation mask of the *same* cell as image with identical size
        :param loc: [C, X, Y, Z] center location of cell
        :param id: unique cell identification number
        :param scores: cell detection likelihood
        :param boxes: [x0, y0, x1, y1] cell detection boxes
        :param cell_type: cell classification ID: 'OHC' or 'IHC'
        :param channel_name: image ordered channel dye names
        """

        self.id = id
        self.loc = loc.cpu()  # [C, X, Y, Z]
        self.frequency = None  # Set by self.calculate_frequency
        self.percent_loc = None  # Set by self.calculate_frequency
        self.type = cell_type  # 'OHC' or 'IHC'
        self.channel_names = channel_name
        self.volume = None
        self.summed = None
        self.distance = None
        self.boxes = box_convert(torch.tensor([self.loc[1], self.loc[2], 30, 30]), 'cxcywh',
                                 'xyxy') if boxes is None else boxes

        self.scores = scores  # only for faster rcnn
        self.channel_stats = None

        self._curve_ind = None
        self._distance_from_curvature = None
        self._distance_is_far_away = False

        if mask is not None and image is not None:
            # Ensure image is in [B, C, X, Y, Z]
            assert image.ndim == 5
            if not image.min() >= 0:
                raise ValueError(image.min())

            # Crop both tensors to the same shape
            image, mask = crop_to_identical_size(image, mask)

            self.volume = mask.gt(0.5).sum().cpu()

            mask = mask[0, 0, ...].gt(0.5).cpu()

            # 0:DAPI
            # 1:GFP
            # 2:MYO7a
            # 3:Actin
            self.summed = image[0, 0, ...][mask].mul(2 ** 12).sum(-1).mean()
            image = image.squeeze(0).reshape(image.shape[1], -1)[:, mask.flatten()].cpu()

            self.channel_stats = {}

            for i in range(image.shape[0]):
                self.channel_stats[channel_name[i]] = _stat(image, i)

    def calculate_frequency(self, curvature: torch.Tensor, distance: torch.Tensor) -> None:
        """
        Calculates cell's best frequency from its place along the cochlear curvature.
        Assigns values to properties: percent_loc, frequency

        Values of greenwood function taken from:
        Moore, B C. (1974). Relation between the critical bandwidth and the frequency-difference limen.
        The Journal of the Acoustical Society of America, 55(2), 359.

        https://en.wikipedia.org/wiki/Greenwood_function

        public double fMouse(double d){ // d is fraction total distance
            //f(Hz) = (10 ^((1-d)*0.92) - 0.680)* 9.8
            return (Math.pow(10, (1-d)*0.92) - 0.680) * 9.8;
        }

        d = d * 100;
        // f(KHz) = (10 ^(((100-d)/100)*2) - 0.4)*200/1000

        Example:
        --------

        >>> from hcat.lib.cell import Cell
        >>> from hcat.lib.functional import PredictCurvature
        >>> import torch
        >>>
        >>> cells = torch.load('array_of_cells.trch')
        >>> masks = torch.load('predicted_segmentation_mask.trch') # torch.shape = [C=1, X, Y, Z]
        >>> curvature, distance, apex = PredictCurvature()(masks)
        >>> for c in cells:
        >>>     c.calculate_frequency(curvature, distance)
        >>> print(f'Best Frequency: {cells[0].frequency}, Hz') # Best Frequency: 1.512 kHz

        :param curvature: 2D curvature array from hcat.lib.functional.PredictCurvature
        :param distance: distance tensor from hcat.lib.functional.PredictCurvature
        :return: None
        """
        # confocal voxel size: 288.88 nm...

        dist = curvature[0, :].sub(self.loc[1]).pow(2) + curvature[1, :].sub(self.loc[2]).pow(2)

        self._curve_ind = torch.argmin(dist)
        self._distance_from_curvature = torch.sqrt(dist[self._curve_ind])
        self._distance_is_far_away = self._distance_from_curvature > 100
        try:
            self.distance = distance[self._curve_ind]
        except:
            print(len(distance), self._curve_ind)
        self.percent_loc = distance[self._curve_ind].div(distance.max())
        self.frequency = (10 ** ((1 - self.percent_loc) * 0.92) - 0.680) * 9.8


if __name__ == '__main__':
    pass



import torch
import torch.nn as nn
from typing import Optional

from hcat.lib.utils import crop_to_identical_size


class jaccard(nn.Module):
    def __init__(self):
        super(jaccard, self).__init__()

    def forward(self, predicted: torch.Tensor, ground_truth: torch.Tensor,
                smooth: float = 1e-10) -> torch.Tensor:
        """
        Returns jaccard index of two torch.Tensors

        :param predicted: [B, I, X, Y, Z] torch.Tensor
                - probabilities calculated from hcat.utils.embedding_to_probability
                  where B: is batch size, I: instances in image
        :param ground_truth: [B, I, X, Y, Z] torch.Tensor
                - segmentation mask for each instance (I).
        :param smooth: float
                - Very small number to ensure numerical stability. Default 1e-10
        :return: jaccard_loss: [1] torch.Tensor
                - Result of Loss Function Calculation
        """

        # Crop both tensors to the same shape
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)

        intersection = (predicted * ground_truth).sum().add(smooth)
        union = (predicted + ground_truth).sum().sub(intersection).add(smooth)

        return 1.0 - (intersection / union)


class dice(nn.Module):
    def __init__(self):
        super(dice, self).__init__()

    def forward(self, predicted: torch.Tensor, ground_truth: torch.Tensor,
                smooth: float = 1e-10) -> torch.Tensor:
        """
        Returns dice index of two torch.Tensors

        :param predicted: [B, I, X, Y, Z] torch.Tensor
                - probabilities calculated from hcat.utils.embedding_to_probability
                  where B: is batch size, I: instances in image
        :param ground_truth: [B, I, X, Y, Z] torch.Tensor
                - segmentation mask for each instance (I).
        :param smooth: float
                - Very small number to ensure numerical stability. Default 1e-10
        :return: dice_loss: [1] torch.Tensor
                - Result of Loss Function Calculation
        """


        # Crop both tensors to the same shape
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)

        intersection = (predicted * ground_truth).sum().add(smooth)
        denominator = (predicted + ground_truth).sum().add(smooth)
        loss = 2 * intersection / denominator

        return 1 - loss


class tversky(nn.Module):
    def __init__(self):
        super(tversky, self).__init__()

    def forward(self, predicted: torch.Tensor, ground_truth: torch.Tensor, smooth: float = 1e-10,
                alpha: float = 0.5, beta: float = 0.5, gamma: float = 0.0, weight: bool = False) -> torch.Tensor:
        """
        Returns dice index of two torch.Tensors

        :param predicted: [B, N, X, Y, Z] torch.Tensor
                - probabilities calculated from hcat.utils.embedding_to_probability
                  where B: is batch size, N: instances in image
        :param ground_truth: [B, N, X, Y, Z] torch.Tensor
                - segmentation mask for each instance (N).
        :param smooth: float
                - Very small number to ensure numerical stability. Default 1e-10
        :param alpha: float
                - Value which penalizes False Positive Values
        :param beta: float
                - Value which penalizes False Negatives
        :param gamma: float
                - Focal loss term
        :return: dice_loss: [1] torch.Tensor
                - Result of Loss Function Calculation
        """

        # Crop both tensors to the same shape
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)

        if gamma != 0:
            # true_positive = (predicted * ground_truth).flatten(start_dim=2).sum(-1)
            # false_positive = (torch.logical_not(ground_truth) * predicted).flatten(start_dim=2).sum(-1).add(1e-10) * alpha
            # false_negative = ((1 - predicted) * ground_truth).flatten(start_dim=2).sum(-1) * beta
            # tversky = (true_positive + smooth) / (true_positive + false_positive + false_negative + smooth)
            # tversky = tversky.add(-1).mul(-1).pow(1 / gamma).mean()
            true_positive = (predicted * ground_truth).sum()
            false_positive = (torch.logical_not(ground_truth) * predicted).sum().add(1e-10) * alpha
            false_negative = ((1 - predicted) * ground_truth).sum() * beta
            tversky = (true_positive + smooth) / (true_positive + false_positive + false_negative + smooth)
            tversky = 1 - tversky
            tversky = tversky.pow(1/gamma)

        else:
            true_positive = (predicted * ground_truth).sum()
            false_positive = (torch.logical_not(ground_truth) * predicted).sum().add(1e-10) * alpha
            false_negative = ((1 - predicted) * ground_truth).sum() * beta
            tversky = (true_positive + smooth) / (true_positive + false_positive + false_negative + smooth)
            tversky = 1 - tversky

        return tversky


class l1(nn.Module):
    def __init__(self):
        super(l1, self).__init__()

    def forward(self, predicted: torch.Tensor, ground_truth: torch.Tensor) -> torch.Tensor:
        """
        Returns mean l1 distance of two torch.Tensors

        :param predicted: [B, I, X, Y, Z] torch.Tensor
                - probabilities calculated from hcat.utils.embedding_to_probability
                  where B: is batch size, I: instances in image
        :param ground_truth: [B, I, X, Y, Z] torch.Tensor
                - segmentation mask for each instance (I).
        :return: l1: [1] torch.Tensor
                - Result of Loss Function Calculation
        """

        # Crop both tensors to the same shape
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)
        loss = torch.nn.L1Loss()

        return loss(predicted, ground_truth)

class L1_CE(nn.Module):
    def __init__(self):
        super(L1_CE, self).__init__()

        self.l1 = torch.nn.L1Loss()
        self.cel = nn.BCEWithLogitsLoss(reduction='none')

    def forward(self, predicted, ground_truth) -> torch.Tensor:
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)

        return self.l1(predicted.float(), ground_truth.float()).mean() + self.cel(predicted.float(), ground_truth.float()).mean()



# class cross_entropy(nn.Module):
#     def __init__(self):
#         super(cross_entropy, self).__init__()
#
#     def forward(self, predicted: torch.Tensor, ground_truth: torch.Tensor) -> torch.Tensor:
#         """
#
#         :param predicted: [B, I, X, Y, Z] torch.Tensor of probabilities calculated from hcat.utils.embedding_to_probability
#                           where B: is batch size, I: instances in image
#         :param ground_truth: [B, I, X, Y, Z] segmentation mask for each instance (I).
#         :return:
#         """
#         raise RuntimeError('Borked. Dont Use. Cuda Error due to not functioning bixel by pixel analysis.')
#
#         # Crop both tensors to the same shape
#         predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)
#
#         loss = torch.nn.CrossEntropyLoss()
#
#         for i in range(ground_truth.shape[1]):
#             ground_truth[:, i, ...] = ground_truth[:, i, ...] * (i + 1)
#
#         ground_truth = ground_truth.sum(1)
#
#         return loss(predicted, ground_truth)

class cross_entropy(nn.Module):
    def __init__(self, method: str = 'pixel', N: Optional[int] = None):
        super(cross_entropy, self).__init__()

        _methods = ['pixel', 'worst_z', 'random']
        if method not in _methods:
            raise ValueError(f'Viable methods for cross entropy loss are {_methods}, not {method}.')

        self.method = method
        self.N = N

        if method == 'random':
            if self.N is None:
                raise ValueError(f'the number of random pixels to draw is not defined. Please set num_random_pixels to a ' +
                                 f'value larger than 1.')
            if self.N <= 1:
                raise ValueError(f'num_random_pixels should be greater than 1 not {self.N}.')

    def forward(self, predicted: torch.Tensor, ground_truth: torch.Tensor, pwl: Optional[torch.Tensor] =None) -> torch.Tensor:
        """
        Pytorch Implementation of Cross Entropy Loss with a pixel by pixel weighting as described in U-NET



        :param pred: torch.Tensor | probability map of shape [B,C,X,Y,Z] predicted by hcat.unet
        :param mask: torch.Tensor | ground truth probability map of shape [B, C, X+dx, Y+dy, Z+dz] that will be cropped
                     to identical size of pred
        :param pwl:  torch.Tensor | weighting map pf shape [B, C, X+dx, Y+dy, Z+dz] that will be cropped to be identical in
                     size to pred and will be used to multiply the cross entropy loss at each pixel.
        :param method: ['pixel', 'worst_z', 'random'] | method by which to weight loss
                            - pixel: weights each pixel size by values in pwl
                            - worst_z: adds additional exponentially decaying weight to z planes on with the largest weight
                                       to the worst performing z plane
                            - random: randomly chooses
        :param num_random_pixels: int or None | Number of randomly selected pixels to draw when method='random'
        :return: torch.float | Average cross entropy loss of all pixels of the predicted mask vs ground truth
        """

        if (ground_truth == 0).sum() == 0:
            raise ValueError(f'There are no background pixels in mask.\n\t(mask==0).sum() == 0 -> True')

        pred_shape = predicted.shape
        n_dim = len(pred_shape)

        is_pwl_none = False
        if pwl is None:
            pwl = torch.ones(predicted.shape, device=predicted.device)
            is_pwl_none = True

        # Crop mask and pwl to the same size as pred
        predicted, ground_truth = crop_to_identical_size(predicted, ground_truth)
        predicted, pwl = crop_to_identical_size(predicted, pwl)
        ground_truth, pwl = crop_to_identical_size(ground_truth, pwl)


        if not is_pwl_none:
            # Hacky way to do this:
            pwl[ground_truth > 0.5] += 2

        cel = nn.BCEWithLogitsLoss(reduction='none')
        loss = None

        if self.method == 'pixel':
            """
            Calculates BCE on ALL pixels of image
            """
            loss = cel(predicted.float(), ground_truth.float())
            loss = (loss * (pwl + 1))

        elif self.method == 'worst_z':
            """
            Calculates BCE on each pixel but weights poor performing z planes more highly
            """
            loss = cel(predicted.float(), ground_truth.float())
            loss = (loss * (pwl + 1))
            scaling = torch.linspace(1, 2, predicted.shape[4]) ** 2
            loss, _ = torch.sort(loss.sum(dim=[0, 1, 2, 3]))
            loss *= scaling.to(loss.device)
            loss /= (predicted.shape[2] * predicted.shape[3])

        elif self.method == 'random':
            """
            Randomly samples an equal number of positive and negative pixels for calculating BCE
            """
            pred = predicted.reshape(-1)
            mask = ground_truth.reshape(-1)

            if (mask == 1).sum() == 0:
                loss = cel(pred.float(), mask.float())
            else:
                pos_ind = torch.randint(low=0, high=int((mask == 1).sum()), size=(1, self.N))[0, :]
                neg_ind = torch.randint(low=0, high=int((mask == 0).sum()), size=(1, self.N))[0, :]

                pred = torch.cat([pred[mask == 1][pos_ind], pred[mask == 0][neg_ind]]).unsqueeze(0)
                mask = torch.cat([mask[mask == 1][pos_ind], mask[mask == 0][neg_ind]]).unsqueeze(0)

                loss = cel(pred.float(), mask.float())

        return loss.mean()


if __name__ == '__main__':
    a = torch.rand((1, 1, 200, 200, 30))
    b = torch.rand((1, 1, 200, 200, 30))
    loss = tversky()
    l = loss((a>0.5).float(), (a > 0.5).long())
    print(l)
